# encoding=utf-8
# %%
import unittest
import warnings

import torch

import torchgdm as tg


class TestDimensionMatching(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print("testing Mie polarizabilities simulation...")

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

        # setup test cases
        self.test_cases = [
            dict(
                wl=700.0,
                radii=[80, 80],
                materials=[16, 16],
                eps_env=1.0,
            ),
            dict(
                wl=900.0,
                radii=[50, 70],
                materials=[12.0 + 0.1j, 9.0],
                eps_env=1.3**2,
            ),
            dict(
                wl=500.0,
                radii=[30, 30],
                materials=[-16.0959 + 0.4438j, -16.0959 + 0.4438j],
                eps_env=1.0,
            ),
            dict(
                wl=600.0,
                radii=[50, 70],
                materials=[-16.0959 + 0.4438j, 16],
                eps_env=1.0,
            ),
            dict(
                wl=650.0,
                radii=[70, 90],
                materials=[9.0, -16.0959 + 0.4438j],
                eps_env=1.5,
            ),
        ]

    def test_calculate_E_H(self):
        import numpy as np

        # try importing external packages
        try:
            # ignore import warnings
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                import pymiecs
        except ModuleNotFoundError:
            if self.verbose:
                print(
                    "Mie tests require package `pymiecs`. "
                    + "Please install via `pip install pymiecs`."
                )
            return
        try:
            # ignore import warnings
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                import treams
        except ModuleNotFoundError:
            if self.verbose:
                print(
                    "Mie tests require package `treams`. "
                    + "Please install via `pip install treams`."
                )
            return

        # - the actual test
        for device in self.devices:
            # ignore warnings temporarily:
            warnings.simplefilter("ignore")
            for conf in self.test_cases:
                wl = conf["wl"]
                k0 = 2 * np.pi / wl
                radii = conf["radii"]
                materials = conf["materials"]
                eps_env = conf["eps_env"]

                # external Mie toolkit
                res_mie = pymiecs.main.Q(
                    tg.to_np(k0),
                    r_core=radii[0],
                    n_core=materials[0] ** 0.5,
                    r_shell=radii[1],
                    n_shell=materials[1] ** 0.5,
                    n_max=1,
                    n_env=tg.to_np(eps_env)**0.5,
                )
                ecs_mie = res_mie["cs_geo"] * res_mie["qext"]
                scs_mie = res_mie["cs_geo"] * res_mie["qsca"]

                # torchGDM using treams-Mie
                env = tg.env.EnvHomogeneous3D(eps_env)
                struct_mie = tg.struct.StructMieSphereEffPola3D(
                    wavelengths=wl, environment=env, radii=radii, materials=materials
                )
                e_inc = tg.env.freespace_3d.PlaneWave()
                sim_tg_mie = tg.Simulation(
                    structures=struct_mie,
                    environment=env,
                    illumination_fields=e_inc,
                    wavelengths=wl,
                    device=device,
                )
                sim_tg_mie.run(verbose=False, progress_bar=False)
                cs_a_mie = sim_tg_mie.get_crosssections(progress_bar=False)

                # test if results match
                np.testing.assert_allclose(
                    ecs_mie[0], tg.to_np(cs_a_mie["ecs"][0]), atol=1e-2, rtol=1e-5
                )
                
                # absorption and thus scattering are not accurate, increase tolerance
                np.testing.assert_allclose(
                    scs_mie[0], tg.to_np(cs_a_mie["scs"][0]), atol=1e-1, rtol=1e-2
                )
                
                if self.verbose:
                    print("  - {}: Mie test passed.".format(device))


# %%
if __name__ == "__main__":
    print("testing Mie polarizabilities...")
    torch.set_printoptions(precision=7)
    unittest.main(argv=["first-arg-is-ignored"], exit=False)



# -*- coding: utf-8 -*-
"""volume discretization for torchgdm
"""
from .pola import StructDiscretizedHexagonal3D, StructDiscretizedCubic3D
from .pola import StructDiscretized3D
from .pola import extract_eff_pola_via_mp_decomposition
from .pola import extract_eff_pola_via_propagation
from . import pola
from . import geometries

from .geometries import discretizer_cubic
from .geometries import discretizer_hexagonalcompact

from .geometries import cube
from .geometries import cuboid
from .geometries import sphere
from .geometries import spheroid
from .geometries import disc
from .geometries import ellipse
from .geometries import split_ring
from .geometries import prism_trigonal

from .geometries import from_image



# -*- coding: utf-8 -*-
"""point polarizability structure class for torchgdm
"""
import warnings
import copy

import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.tools.misc import get_default_device
from torchgdm.struct.base_classes import StructBase
from torchgdm.tools import interp
from torchgdm.tools.geometry import rotation_x, rotation_y, rotation_z
from torchgdm.tools.geometry import test_structure_distances
from torchgdm.tools.geometry import get_enclosing_sphere_radius
from torchgdm.tools.misc import ptp


# --- base class volume discretized structure container - 3D
class StructEffPola3D(StructBase):
    """base class 3D point polarizability structure

    Defines a basic effective polarizability based structure
    """

    __name__ = "effective point polarizability (3D) structure class"

    def __init__(
        self,
        positions: torch.Tensor,
        alpha_dicts: list,
        device: torch.device = None,
        environment=None,
        shift_z_to_r0: bool = True,
    ):
        """3D point polarizability class

        The main information is provided in the `alpha_dicts`, which is a list of dicts with the full effective polarizability definitions. Each dict defines one structure and must contain following:
            - 'wavelengths': wavelengths at which the polarizabilities are calculated
            - 'alpha_6x6':
                6x6 polarizability tensors of shape [len(wavelengths),6,6]. Coupling terms:
                    - pE: [:, :3, :3]
                    - mH: [:, 3:, 3:]
                    - mE: [:, 3:, :3]
                    - pH: [:, :3, 3:]
            optional keys:
            - 'full_geometry': the original volume discretization of the represented geometry
            - 'r0': the origin of the effective polarizabilities with respect to optional 'full_geometry'
            - 'enclosing_radius': enclosing radius of the original structure

        Args:
            positions (torch.Tensor): polarizability positions
            alpha_dicts (list): list of polarizability model dictionaries
            device (torch.device, optional): Defaults to "cpu".
            environment (_type_, optional): 3D environment class. Defaults to None.
            shift_z_to_r0 (bool, optional): If True, if a position z-value is zero, each polarizability model's z position will be shifted to the height of the effective dipole development center. Defaults to True.

        Raises:
            ValueError: _description_
        """
        super().__init__(device=device)
        self.n_dim = 3

        # expand positions, put single scatterer in list
        self.positions = torch.as_tensor(
            positions, dtype=DTYPE_FLOAT, device=self.device
        )
        if len(self.positions.shape) == 1:
            assert len(self.positions) == 3
            self.positions = self.positions.unsqueeze(0)

        # environment at which alpha has been extracted (if given):
        self.environment = environment

        # single alpha_dict: put in list
        if type(alpha_dicts) == dict:
            alpha_dicts = [alpha_dicts] * len(self.positions)
        assert type(alpha_dicts[0]) == dict

        for _adict in alpha_dicts:
            assert "wavelengths" in _adict
            if "alpha_6x6" not in _adict:
                raise ValueError(
                    "Polarizability-description dicts must contain polarizaility tensors "
                    + "under the dict key: 'alpha_6x6'."
                )

        # use first pola-tensor wavelength range
        wavelengths = alpha_dicts[0]["wavelengths"]
        self.wavelengths_data = torch.as_tensor(
            wavelengths, dtype=DTYPE_FLOAT, device=self.device
        )

        for _adict in alpha_dicts:
            _wls = torch.as_tensor(
                _adict["wavelengths"], dtype=DTYPE_FLOAT, device=self.device
            )
            if not torch.all(torch.isin(self.wavelengths_data, _wls)):
                warnings.warn(
                    "Pre-calculated wavelengths for the structures are not identical. "
                    + "Try to use linear interpolation to fill missing values. "
                    + "This may be inaccurate."
                )
            if (torch.min(_wls) > torch.min(self.wavelengths_data)) or (
                torch.max(_wls) < torch.max(self.wavelengths_data)
            ):
                raise ValueError(
                    "Interpolation not possible. Pre-calculated wavelengths of the structures "
                    + "must be within the same range to allow interpolation."
                )

        # optionally shift z such that 'bottom' is at z=0. Do only if z=0
        if shift_z_to_r0:
            for i, p in enumerate(self.positions):
                if p[2] == 0:
                    p[2] = torch.as_tensor(
                        alpha_dicts[i]["r0"][2], dtype=p.dtype, device=p.device
                    )

        # populate polarizability tensors database for each position at
        # pre-calculated wavelengths. if necessary, interpolate between data points
        alpha_data = []
        enclosing_radius = []
        full_geometries = []
        for i, _adict in enumerate(alpha_dicts):
            alpha_data.append(
                self._interpolate_single_alpha(
                    wls=self.wavelengths_data,
                    a_data=_adict["alpha_6x6"],
                    wl_data=_adict["wavelengths"],
                )
            )

            # optional values
            if "full_geometry" in _adict:
                full_geometries.append(
                    torch.as_tensor(
                        _adict["full_geometry"] - _adict["r0"],
                        dtype=DTYPE_FLOAT,
                        device=self.device,
                    )
                    + self.positions[i]
                )

            if "enclosing_radius" in _adict:
                enclosing_radius.append(_adict["enclosing_radius"] * 2)
            else:
                if "full_geometry" in _adict:
                    _r_eff = get_enclosing_sphere_radius(full_geometries[-1])
                    enclosing_radius.append(_r_eff)
                else:
                    enclosing_radius.append(0)

        self.full_geometries = full_geometries

        # define zeros for later reuse
        self.zeros_spectral = torch.zeros(
            (len(self.positions), len(self.wavelengths_data), 6, 6),
            dtype=DTYPE_COMPLEX,
            device=self.device,
        )

        # effective polarizabilities of each meshpoint at each wavelength: shape (Npos, Nwl, 6, 6)
        if len(alpha_data) != 0:
            self.alpha_data = torch.stack(alpha_data)
        else:
            self.alpha_data = self.zeros_spectral

        # case of a single point-polarizability: expand dimensions
        if len(self.alpha_data.shape) == 3:
            self.alpha_data = self.alpha_data.unsqueeze(0)

        # selfterms are zero.
        self.selfterms_data = self.zeros_spectral

        # populate lookup tables
        self.create_lookup()

        # other parameters ("step" corresponds to the effective diameter)
        self.step = torch.as_tensor(
            enclosing_radius, dtype=DTYPE_FLOAT, device=self.device
        )

        # center of gravity of ensemble of all points
        self.r0 = self.get_center_of_mass()

        # verify dimensions
        if self.alpha_data.shape != self.zeros_spectral.shape:
            raise ValueError(
                "Wrong shape of a polarizability tensor list. "
                + "Needs to be (N_dipoles, N_wavelengths, 6, 6)."
            )

        # effective EE / HH polarizabilities
        self.evaluation_terms = ["E", "H"]  # possible terms 'E' and 'H'

    def __repr__(self, verbose=False):
        """description about structure"""
        out_str = ""
        out_str += "------ 3D effective ED / MD polarizability nano-object -------"
        out_str += "\n" + " nr. of dipole-pairs:    {}".format(
            len(self.get_all_positions())
        )
        out_str += "\n" + " nominal enclosing sphere diameters (nm): {}".format(
            [round(float(f), 1) for f in torch.unique(self.step)]
        )
        if len(self.full_geometries) > 0:
            pos = torch.cat(self.full_geometries)
            out_str += "\n" + " original geometry: "
            out_str += "\n" + "  - replacing nr. of meshpoints: {}".format(len(pos))
            bnds = ptp(pos, dim=0)
            out_str += "\n" + "  - size & position:"
            out_str += "\n" + "        X-extension    :    {:.1f} (nm)".format(bnds[0])
            out_str += "\n" + "        Y-extension    :    {:.1f} (nm)".format(bnds[1])
            out_str += "\n" + "        Z-extension    :    {:.1f} (nm)".format(bnds[2])
            out_str += "\n" + "  - center of mass :    ({:.1f}, {:.1f}, {:.1f})".format(
                *[float(f) for f in self.get_center_of_mass()]
            )

        return out_str

    def create_lookup(self):
        """Create a lookup table for the polarizability tensors"""
        # populate lookup tables with pre-calculated data
        self.lookup = {}
        for i_wl, wl in enumerate(self.wavelengths_data):
            self.lookup[float(wl)] = self.alpha_data[:, i_wl, :, :]

    def set_device(self, device):
        """move all tensors of the class to device"""
        super().set_device(device)
        if self.environment is not None:
            self.environment.set_device(device)

        self.positions = self.positions.to(device)
        if len(self.full_geometries) > 0:
            for i_g in range(len(self.full_geometries)):
                self.full_geometries[i_g] = self.full_geometries[i_g].to(device)

        self.zeros_spectral = self.zeros_spectral.to(device)
        self.step = self.step.to(device)
        self.r0 = self.r0.to(device)
        self.wavelengths_data = self.wavelengths_data.to(device)

        self.alpha_data = self.alpha_data.to(device)
        self.selfterms_data = self.selfterms_data.to(device)

        # transfer the lookup tables
        for wl in self.lookup:
            self.lookup[wl] = self.lookup[wl].to(device)

    def _interpolate_single_alpha(self, wls, a_data, wl_data):
        # convert to tensor
        wls = torch.as_tensor(wls, dtype=DTYPE_FLOAT, device=self.device)
        wl_dat = torch.as_tensor(wl_data, dtype=DTYPE_FLOAT, device=self.device)
        a_dat = torch.as_tensor(a_data, dtype=DTYPE_COMPLEX, device=self.device)

        # wavelength-interpolation of each tensor component
        a_ip = torch.zeros((len(wls), 6, 6), dtype=DTYPE_COMPLEX, device=self.device)
        for i in range(6):
            for j in range(6):
                _func_ip = interp.RegularGridInterpolator([wl_dat], a_dat[:, i, j])
                _ip = _func_ip([wls])
                a_ip[:, i, j] = _ip

        return a_ip

    def interpolate_alpha(self, wls, a_data_many, wl_data, lookup=None):
        """interpolate the polarizabilities between available wavelengths"""
        a_ip = torch.zeros(
            (len(a_data_many), len(wls), 6, 6), dtype=DTYPE_COMPLEX, device=self.device
        )

        # iterate all polarizabilities (different structures)
        for i_a, a_data in enumerate(a_data_many):
            _a = self._interpolate_single_alpha(wls, a_data, wl_data)
            a_ip[i_a] = _a

        # optionally add wavelengths to lookup
        if lookup is not None:
            for i_wl, wl in enumerate(wls):
                wl = float(wl)
                if wl not in lookup:
                    lookup[wl] = a_ip[:, i_wl, :, :]

        return a_ip

    # --- self-terms
    # self-terms are zero
    # TODO: Make them user-definable (use same approach as for alpha_XX)

    # --- polarizabilities
    def _call_interpolation(self, wavelength):
        warnings.warn(
            "Interpolating polarizabilities at wavelength {:.3f}.".format(
                float(wavelength)
            )
        )
        self.interpolate_alpha(
            [wavelength],
            self.alpha_data,
            self.wavelengths_data,
            lookup=self.lookup,
        )

    def get_polarizability_pE(self, wavelength: float, environment) -> torch.Tensor:
        """return pE polarizability tensors (3x3) at each position

        Args:
            wavelength (float): in nm
            environment (environment class): 3D environement class

        Returns:
            torch.Tensor: pE polarizability tensor
        """
        if float(wavelength) not in self.lookup:
            self._call_interpolation(wavelength)

        return self.lookup[float(wavelength)][:, :3, :3]

    def get_polarizability_mE(self, wavelength: float, environment) -> torch.Tensor:
        """return mE polarizability tensors (3x3) at each position

        Args:
            wavelength (float): in nm
            environment (environment class): 3D environement class

        Returns:
            torch.Tensor: mE polarizability tensor
        """
        if float(wavelength) not in self.lookup:
            self._call_interpolation(wavelength)

        return self.lookup[float(wavelength)][:, 3:, :3]

    def get_polarizability_pH(self, wavelength: float, environment) -> torch.Tensor:
        """return pH polarizability tensors (3x3) at each position

        Args:
            wavelength (float): in nm
            environment (environment class): 3D environement class

        Returns:
            torch.Tensor: pH polarizability tensor
        """
        if float(wavelength) not in self.lookup:
            self._call_interpolation(wavelength)

        return self.lookup[float(wavelength)][:, :3, 3:]

    def get_polarizability_mH(self, wavelength: float, environment) -> torch.Tensor:
        """return mH polarizability tensors (3x3) at each position

        Args:
            wavelength (float): in nm
            environment (environment class): 3D environement class

        Returns:
            torch.Tensor: mH polarizability tensor
        """
        if float(wavelength) not in self.lookup:
            self._call_interpolation(wavelength)

        return self.lookup[float(wavelength)][:, 3:, 3:]

    # --- plotting
    def plot(
        self,
        projection="auto",
        scale=1.0,
        color="auto",
        linestyle_circle=(0, (2, 2)),
        color_circle="auto",
        color_circle_fill=None,
        alpha=1,
        show_grid=True,
        color_grid="auto",
        alpha_grid=0.25,
        legend=True,
        set_ax_aspect=True,
        reset_color_cycle=True,
        **kwargs,
    ):
        """plot the point polarizability structure (2D)

        Args:
            projection (str, optional): Cartesian projection. Default: "XY" or plane in which all dipoles lie. Defaults to "auto".
            scale (float, optional): scaling factor of the grid cells, if shown. Defaults to 1.0.
            color (str, optional): plot color. Defaults to "auto".
            linestyle_circle (tuple, optional): optional line style for enclosing circle. Defaults to (0, (2, 2)).
            color_circle (str, optional): optional alternative color for enclosing circle. Defaults to "auto".
            color_circle_fill (_type_, optional): optional alternative fill color for enclosing circle. Defaults to None.
            alpha (int, optional): optional transparency. Defaults to 1.
            show_grid (bool, optional): whether to show mesh grid (if available in structure). Defaults to True.
            color_grid (str, optional): optional alternative color for the mesh grid. Defaults to "auto".
            alpha_grid (float, optional): optional alternative transparency for the mesh grid. Defaults to 0.25.
            legend (bool, optional): show legend. Defaults to True.
            set_ax_aspect (bool, optional): automatically set aspect ratio to equal. Defaults to True.
            reset_color_cycle (bool, optional): reset color cycle after finishing the plot. Defaults to True.

        Returns:
            matplotlib axes
        """
        from torchgdm.visu import visu2d

        im = visu2d.geo2d._plot_structure_eff_pola(
            self,
            projection=projection,
            scale=scale,
            color=color,
            linestyle_circle=linestyle_circle,
            color_circle=color_circle,
            color_circle_fill=color_circle_fill,
            alpha=alpha,
            show_grid=show_grid,
            color_grid=color_grid,
            alpha_grid=alpha_grid,
            legend=legend,
            set_ax_aspect=set_ax_aspect,
            reset_color_cycle=reset_color_cycle,
            **kwargs,
        )
        return im

    def plot_contour(
        self,
        projection="auto",
        color="auto",
        set_ax_aspect=True,
        alpha=1.0,
        alpha_value=None,
        reset_color_cycle=True,
        **kwargs,
    ):
        """plot the contour of the underlying mesh (2D)

        Args:
            projection (str, optional): which cartesian plane to project onto. Defaults to "auto".
            color (str, optional): optional matplotlib compatible color. Defaults to "auto".
            set_ax_aspect (bool, optional): If True, will set aspect of plot to equal. Defaults to True.
            alpha (float, optional): matplotlib transparency value. Defaults to 1.0.
            alpha_value (float, optional): alphashape value. If `None`, try to automatically optimize for best enclosing of structure. Defaults to None.
            reset_color_cycle (bool, optional): reset color cycle after finishing the plot. Defaults to True.

        Returns:
            matplotlib line: matplotlib's `scatter` output
        """
        from torchgdm.visu import visu2d

        im = visu2d.geo2d.contour(
            self,
            projection=projection,
            color=color,
            set_ax_aspect=set_ax_aspect,
            alpha=alpha,
            alpha_value=alpha_value,
            reset_color_cycle=reset_color_cycle,
            **kwargs,
        )
        return im

    def plot3d(self, **kwargs):
        """plot the point polarizability structure (3D)"""
        from torchgdm.visu import visu3d

        return visu3d.geo3d._plot_structure_eff_3dpola(self, **kwargs)

    # --- geometry operations
    def translate(self, vector):
        """return a copy moved by `vector`"""
        vector = torch.as_tensor(vector, dtype=DTYPE_FLOAT, device=self.device)

        _shifted = self.copy()

        _shifted.positions += vector.unsqueeze(0)
        _shifted.r0 = _shifted.get_center_of_mass()

        if len(_shifted.full_geometries) > 0:
            for _g in _shifted.full_geometries:
                _g += vector.unsqueeze(0)

        return _shifted

    def set_center_of_mass(self, r0_new: torch.Tensor):
        """move center of mass to new position `r0_new` (in-place)"""
        r0_new = torch.as_tensor(r0_new, device=self.device)

        if len(r0_new.shape) != 1:
            if len(r0_new) not in [2, 3]:
                raise ValueError("`r0_new` needs to be (X,Y) or (X,Y,Z) tuple.")
        r0_old = self.get_center_of_mass()

        # treat case r0_new is 2D (X,Y)
        if len(r0_new) == 2:
            r0_new = torch.as_tensor(
                [r0_new[0], r0_new[1], r0_old[2]], device=self.device
            )

        # move to origin, then to new location
        self.positions -= r0_old - r0_new

        if len(self.full_geometries) > 0:
            for _g in self.full_geometries:
                # move to origin, the to new pos.
                _g -= r0_old - r0_new

        self.r0 = self.get_center_of_mass()

    def rotate(self, alpha, center=torch.as_tensor([0.0, 0.0, 0.0]), axis="z"):
        """rotate the structure

        Args:
            alpha (float): rotation angle (in rad)
            center (torch.Tensor, optional): center of rotation axis. Defaults to torch.as_tensor([0.0, 0.0, 0.0]).
            axis (str, optional): rotation axis, one of ['x', 'y', 'z']. Defaults to 'z'.

        Raises:
            ValueError: unknown rotation axis

        Returns:
            :class:`StructEffPola3D`: copy of structure with rotated geometry
        """
        _s_rot = self.copy()
        center = center.to(dtype=DTYPE_FLOAT, device=self.device)

        if axis.lower() == "x":
            rot = rotation_x(alpha, device=self.device)
        elif axis.lower() == "y":
            rot = rotation_y(alpha, device=self.device)
        elif axis.lower() == "z":
            rot = rotation_z(alpha, device=self.device)
        else:
            raise ValueError("Unknown rotation axis ''.".format(axis))

        # rotate positions (if multiple)
        if len(_s_rot.positions) > 1:
            _s_rot.positions = torch.matmul(_s_rot.positions - center, rot) + center

        # rotate full discretizations (if available)
        for i_g, _geo in enumerate(_s_rot.full_geometries):
            if len(_geo) > 1:
                _s_rot.full_geometries[i_g] = torch.matmul(
                    _geo - (center + self.r0), rot
                ) + (center + self.r0)

        # rotate polarizability sub-tensors
        rot = rot.to(DTYPE_COMPLEX).unsqueeze(0).unsqueeze(0)
        rotT = rot.transpose(-2, -1)
        for m in [0, 3]:
            for n in [0, 3]:
                _s_rot.alpha_data[..., m : m + 3, n : n + 3] = torch.matmul(
                    torch.matmul(
                        rotT,
                        _s_rot.alpha_data[..., m : m + 3, n : n + 3],
                    ),
                    rot,
                )

        # update lookup tables
        _s_rot.create_lookup()

        return _s_rot

    def combine(
        self, other, inplace=False, refresh_lookup=True, on_distance_violation="warn"
    ):
        """combine with a second structure

        Structures must be of same coupling type (electric / magnetic)

        Args:
            other (_type_): _description_
            inplace (bool, optional): Don't copy original structure, just add other structure. Can be necessary e.g. when gradients are required. Defaults to False.
            refresh_lookup (bool, optional): refresh the polarizability lookup table. Defaults to True.
            on_distance_violation (str, optional): can be "error", "warn", None (do nothing). Defaults to "error".

        Returns:
            :class:`StructBase`: new structure
        """
        if inplace:
            new_struct = self
        else:
            new_struct = self.copy()

        assert torch.all(new_struct.wavelengths_data == other.wavelengths_data)
        assert type(self) == type(other)

        N_dist1, N_dist2 = test_structure_distances(self, other)
        if on_distance_violation == "error" and (N_dist1 + N_dist2 > 0):
            raise ValueError(
                "Several meshpoints in structures are too close (struct1: {}, structu2: {})!".format(
                    N_dist1, N_dist2
                )
            )
        elif on_distance_violation == "warn" and (N_dist1 + N_dist2 > 0):
            warnings.warn(
                "Several meshpoints in structures are too close (struct1: {}, structu2: {})!".format(
                    N_dist1, N_dist2
                )
            )

        new_struct.positions = torch.concatenate(
            [new_struct.positions, other.positions], dim=0
        )

        new_struct.zeros_spectral = torch.concatenate(
            [new_struct.zeros_spectral, other.zeros_spectral], dim=0
        )

        new_struct.step = torch.concatenate([new_struct.step, other.step], dim=0)
        new_struct.r0 = new_struct.get_center_of_mass()

        new_struct.alpha_data = torch.concatenate(
            [new_struct.alpha_data, other.alpha_data], dim=0
        )

        new_struct.selfterms_data = torch.concatenate(
            [new_struct.selfterms_data, other.selfterms_data], dim=0
        )

        # finally, add full geometries
        new_struct.full_geometries = new_struct.full_geometries + other.full_geometries

        # create lookup
        if refresh_lookup:
            new_struct.create_lookup()

        return new_struct


# --- Mie sphere
class StructMieSphereEffPola3D(StructEffPola3D):
    """class for Mie-theory based 3D point polarizability

    Requires external package `treams`
    !!! Does not support automatic differentiation !!!

    Defines a point polarizability representing a sphere using
    first order (dipolar) Mie coefficients
    """

    __name__ = "Mie-theory sphere dipolar polarizability (3D) structure class"

    def __init__(
        self,
        wavelengths: torch.Tensor,
        radii: list,
        materials: list,
        environment=None,
        r0: torch.Tensor = None,
        device: torch.device = None,
        quadrupol_tol=0.15,
    ):
        """3D point polarizability class for a core-shell sphere (Mie)

        Use Mie theory in dipole approximation (first order) the get an
        effective polarizability model for a core-shell sphere.
        Requires the `treams` package for Mie coefficient calculation.
        https://github.com/tfp-photonics/treams

        Args:
            wavelengths (torch.Tensor): list of wavelengths to evaluate (nm)
            radii (list): list of the sphere's core and (multiple) shell radii (in nm).
            materials (list): materials of core and shell(s). A float or int is interpreted as permittivity value.
            environment (environment instance, optional): Homogeneous 3D environment to evaluate Mie theory in. Defaults to None, which uses vacuum.
            r0 (torch.Tensor, optional): polarizability position (x,y,z). If not given, is set to (0, 0, r_sphere). Defaults to None
            alpha_dicts (list): list of polarizability model dictionaries
            device (torch.device, optional): Defaults to "cpu".
            quadrupol_tol (float, optional): ratio of tolerable residual quadrupole terms relative to the dipole order before warning. Defaults to 0.15.

        Raises:
            ValueError: incorrect parameters
        """
        # prep and imports
        from torchgdm.tools.misc import to_np
        from torchgdm.materials.base_classes import MaterialBase
        from torchgdm.materials import MatConstant
        import numpy as np

        # TODO: Replace with differentiable Mie code
        try:
            # ignore import warnings
            with warnings.catch_warnings():
                import treams
        except ModuleNotFoundError:
            print(
                "Mie polarizabilities require package `treams`. "
                + "Please install via `pip install treams`."
            )
            raise

        if device is None:
            self.device = get_default_device()
        else:
            self.device = device

        # environment
        if environment is None:
            from torchgdm.env import EnvHomogeneous3D

            env = EnvHomogeneous3D(device=self.device)
            warnings.warn("No environment specified. Using a vacuum environment.")
        else:
            env = environment

        if env.n_dim != 3:
            raise ValueError(
                f"Mie polarizability requires a 3D environement (is {env.n_dim}D). "
                + "Please use a compatible environment."
            )

        # tensor conversion
        wavelengths = torch.as_tensor(wavelengths, dtype=DTYPE_FLOAT, device=device)
        wavelengths = torch.atleast_1d(wavelengths)
        k0 = 2 * torch.pi / wavelengths
        k0 = torch.as_tensor(k0, device=self.device)

        # radii to array
        radii = np.atleast_1d(radii)
        r_enclosing = np.max(radii)  # outer radius

        # if single material, put in list
        if not hasattr(materials, "__iter__"):
            materials = [materials]

        # main Mie extraction and setup
        a_n = np.zeros((2, len(wavelengths)), dtype=np.complex128)
        b_n = np.zeros_like(a_n)
        for i_wl, wl in enumerate(wavelengths):
            # embedding medium and wavevector therein
            n_env = to_np(env.env_material.get_epsilon(wavelength=wl))[0, 0] ** 0.5

            # core and shell materials
            mat_treams = []
            for mat in materials:
                if type(mat) in [float, int, complex]:
                    mat = MatConstant(mat)
                n_mat = to_np(mat.get_epsilon(wavelength=wl))[0, 0] ** 0.5
                mat_treams.append(treams.Material(n_mat**2))

            # add environment material last
            mat_treams.append(treams.Material(n_env**2))

            for l in [1, 2]:
                miecoef = treams.coeffs.mie(
                    l, to_np(k0[i_wl]) * np.array(radii), *zip(*mat_treams)
                )
                a_n[l - 1, i_wl] = -miecoef[0, 0] - miecoef[0, 1]
                b_n[l - 1, i_wl] = -miecoef[0, 0] + miecoef[0, 1]

        a_n = torch.as_tensor(a_n, device=self.device)
        b_n = torch.as_tensor(b_n, device=self.device)

        # check if dipole approximation is good
        a_quadrupol_tol = a_n[1].abs() / a_n[0].abs()
        wls_violation_a = to_np(wavelengths[a_quadrupol_tol.to("cpu") > quadrupol_tol])
        if len(wls_violation_a) > 0:
            warnings.warn(
                "Mie series: {} wavelengths with ".format(len(wls_violation_a))
                + "significant residual electric quadrupole contribution: "
                + "{} nm".format([round(r, 1) for r in wls_violation_a])
            )

        b_quadrupol_tol = b_n[1].abs() / b_n[0].abs()
        wls_violation_b = to_np(wavelengths[b_quadrupol_tol.to("cpu") > quadrupol_tol])
        if len(wls_violation_b) > 0:
            warnings.warn(
                "Mie series: {} wavelengths with ".format(len(wls_violation_b))
                + "significant residual magnetic quadrupole contribution: "
                + "{} nm".format([round(r, 1) for r in wls_violation_b])
            )

        # convert to polarizabilities (units of volume)
        # see: García-Etxarri, A. et al. Optics Express 19, 4815 (2011)
        a_pE = 1j * 3 / 2 * a_n[0] / k0**3 / n_env**1
        a_mH = 1j * 3 / 2 * b_n[0] / k0**3 / n_env**2

        # populate 6x6 polarizabilities for all wavelengths
        alpha_6x6 = torch.zeros(
            (len(wavelengths), 6, 6), dtype=DTYPE_COMPLEX, device=self.device
        )
        alpha_6x6[:, torch.arange(3), torch.arange(3)] += a_pE.unsqueeze(1)
        alpha_6x6[:, torch.arange(3, 6), torch.arange(3, 6)] += a_mH.unsqueeze(1)

        # set center of mass
        if r0 is None:
            r0 = torch.as_tensor([0.0, 0.0, 0.0], dtype=DTYPE_FLOAT, device=self.device)
        else:
            r0 = torch.as_tensor(r0, dtype=DTYPE_FLOAT, device=self.device)
            r0 = r0.squeeze()
            assert len(r0) == 3

        # wrap up in a dictionary compatible with the point dipole structure class
        alpha_dict = dict(
            r0=r0,
            r0_MD=r0,
            r0_ED=r0,
            alpha_6x6=alpha_6x6,
            wavelengths=wavelengths,
            enclosing_radius=r_enclosing,
            k0_spectrum=k0,
        )

        # - point polarizability structure with Mie dipolar response
        super().__init__(
            positions=r0,
            alpha_dicts=[alpha_dict],
            environment=env,
            device=self.device,
        )



# encoding=utf-8
# %%
import unittest

import torch

import torchgdm as tg


class TestSim3DVacuumPlanWave(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print("testing 3D discretized simulation, plane wave illumination...")

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

        # --- setup a test case
        # - environment
        eps_env = 1.0
        mat_env = tg.materials.MatConstant(eps_env)
        env = tg.env.freespace_3d.EnvHomogeneous3D(env_material=mat_env)

        # - illumination field(s)
        self.wl = 550.0
        wavelengths = torch.tensor([self.wl])
        e_inc_list = [
            tg.env.freespace_3d.PlaneWave(e0p=1.0, e0s=0.6, inc_angle=torch.pi),
        ]

        # - structure
        step = 25.0
        mat_struct = tg.materials.MatConstant(eps=10.0)
        struct1 = tg.struct.StructDiscretizedCubic3D(
            tg.struct.volume.cuboid(l=3, w=2, h=4),
            step,
            mat_struct,
            radiative_correction=False,
        )

        self.sim = tg.simulation.Simulation(
            structures=[struct1],
            environment=env,
            illumination_fields=e_inc_list,
            wavelengths=wavelengths,
        )

    def test_calculate_E_H(self):
        for device in self.devices:
            self.sim.set_device(device)
            self.sim.run(calc_missing=True, verbose=False, progress_bar=False)

            # compare random e-field
            E_test = self.sim.fields_inside[self.wl].efield[0][7]
            # print(E_test)
            E_truth = torch.as_tensor(
                [
                    0.13236962 - 0.2889568j,
                    0.04079278 - 0.11384597j,
                    -0.06021021 - 0.00766215j,
                ],
                dtype=tg.constants.DTYPE_COMPLEX,
                device=device,
            )
            torch.testing.assert_close(E_test, E_truth, rtol=1e-5, atol=0)

            # compare random h-field
            H_test = self.sim.fields_inside[self.wl].hfield[0][2]
            # print(H_test)
            H_truth = torch.as_tensor(
                [
                    0.49382266 - 0.44141245j,
                    -0.8956885 + 0.8145405j,
                    0.05164324 + 0.07042729j,
                ],
                dtype=tg.constants.DTYPE_COMPLEX,
                device=device,
            )
            torch.testing.assert_close(H_test, H_truth, rtol=1e-5, atol=0)
            if self.verbose:
                print("  - {}: 3D sim E, H field test passed.".format(device))


class TestSim2DVacuumPlanWave(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print("testing 2D discretized simulation, plane wave illumination...")

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

        # --- setup a test case
        # - environment
        eps_env = 2.25
        mat_env = tg.materials.MatConstant(eps_env)
        env = tg.env.freespace_2d.EnvHomogeneous2D(env_material=mat_env)

        # - illumination field(s)
        self.wl = 550.0
        wavelengths = torch.tensor([self.wl])
        e_inc_list = [
            tg.env.freespace_2d.PlaneWave(e0p=1.0, e0s=0.6, inc_angle=torch.pi / 3),
        ]

        # - structure
        step = 25.0
        mat_struct = tg.materials.MatConstant(eps=6.0)
        struct1 = tg.struct.StructDiscretizedSquare2D(
            tg.struct.surface_2d.rectangle(l=3, h=2), step, mat_struct
        )

        self.sim = tg.simulation.Simulation(
            structures=[struct1],
            environment=env,
            illumination_fields=e_inc_list,
            wavelengths=wavelengths,
        )

    def test_calculate_E_H(self):
        for device in self.devices:
            self.sim.set_device(device)
            self.sim.run(calc_missing=True, verbose=False, progress_bar=False)

            # compare random e-field
            E_test = self.sim.fields_inside[self.wl].efield[0][3]
            # print(E_test)
            E_truth = torch.as_tensor(
                [
                    -0.2927338 - 0.2186406j,
                    0.5191928 + 0.5264899j,
                    -0.4140966 - 0.1886911j,
                ],
                dtype=tg.constants.DTYPE_COMPLEX,
                device=device,
            )
            torch.testing.assert_close(E_test, E_truth, rtol=1e-2, atol=1e-5)

            # compare random h-field
            H_test = self.sim.fields_inside[self.wl].hfield[0][4]
            # print(H_test)
            H_truth = torch.as_tensor(
                [
                    -0.5955392 + 0.2554660j,
                    -1.6774530 + 0.6361306j,
                    -1.0642812 + 0.3601927j,
                ],
                dtype=tg.constants.DTYPE_COMPLEX,
                device=device,
            )
            torch.testing.assert_close(H_test, H_truth, rtol=1e-2, atol=1e-5)
            if self.verbose:
                print("  - {}: 2D sim E, H field test passed.".format(device))


# %%
if __name__ == "__main__":
    print("testing discretization simulation...")
    torch.set_printoptions(precision=7)
    unittest.main(argv=["first-arg-is-ignored"], exit=False)



# -*- coding: utf-8 -*-
"""
exact multipole decomposition post-processing tools for torchgdm
"""
# %%
import warnings

import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.tools.misc import get_default_device


# --- double factorial via recursion (used as fixed value below)
def _doublefactorial(n):
    if n <= 0:
        return 1
    else:
        return n * _doublefactorial(n - 2)


# --- first 4 spherical Bessel functions in pytorch
def sph_j0(x: torch.Tensor, asymptotic_threshold=0.001):
    """spherical Bessel function of zero order

    for numerical stability, use asymptotic solutions at small arguments.

    Args:
        x (torch.Tensor): argument of Bessel function
        asymptotic_threshold (float, optional): threshold below which to use small argument asymptotic approximation. Defaults to 0.001.

    Returns:
        float: bessel result
    """
    x = torch.as_tensor(x, dtype=DTYPE_COMPLEX)
    j0 = torch.where(torch.abs(x) > asymptotic_threshold, torch.sin(x) / x, 1)
    return j0


def sph_j1(x: torch.Tensor, asymptotic_threshold=0.01):
    """spherical Bessel function of first order

    for numerical stability, use asymptotic solutions at small arguments.

    Args:
        x (torch.Tensor): argument of Bessel function
        asymptotic_threshold (float, optional): threshold below which to use small argument asymptotic approximation. Defaults to 0.01.

    Returns:
        float: bessel result
    """
    x = torch.as_tensor(x, dtype=DTYPE_COMPLEX)
    _sin_term = torch.sin(x) / x**2
    _cos_term = torch.cos(x) / x
    j1 = torch.where(torch.abs(x) > asymptotic_threshold, _sin_term - _cos_term, x / 3)
    return j1


def sph_j2(x: torch.Tensor, asymptotic_threshold=0.1):
    """spherical Bessel function of second order

    for numerical stability, use asymptotic solutions at small arguments.

    Args:
        x (torch.Tensor): argument of Bessel function
        asymptotic_threshold (float, optional): threshold below which to use small argument asymptotic approximation. Defaults to 0.1.

    Returns:
        float: bessel result
    """
    x = torch.as_tensor(x, dtype=DTYPE_COMPLEX)
    _sin_term = (3 / x**2 - 1) * (torch.sin(x) / x)
    _cos_term = 3 * torch.cos(x) / x**2
    j2 = torch.where(
        torch.abs(x) > asymptotic_threshold, _sin_term - _cos_term, x**2 / 15
    )
    return j2


def sph_j3(x: torch.Tensor, asymptotic_threshold=0.5):
    """spherical Bessel function of third order

    for numerical stability, use asymptotic solutions at small arguments.

    Args:
        x (torch.Tensor): argument of Bessel function
        asymptotic_threshold (float, optional): threshold below which to use small argument asymptotic approximation. Defaults to 0.5.

    Returns:
        float: bessel result
    """
    x = torch.as_tensor(x, dtype=DTYPE_COMPLEX)
    _sin_term = (15 / x**3 - 6 / x) * (torch.sin(x) / x)
    _cos_term = (15 / x**2 - 1) * (torch.cos(x) / x)
    j3 = torch.where(
        torch.abs(x) > asymptotic_threshold, _sin_term - _cos_term, x**3 / 105
    )
    return j3


# Bessel functions of first kind
def Jn(n: int, z: torch.Tensor):
    """integer order Bessel functions of first kind via recurrence formula

    Notes:
        - For pure real arguments, use recurrence: J_n+1 = (2n/z) J_n - J_n-1
        - for purely imaginary arguments, use identity via modif. Bessel of first kind
          Jn(1j*z) = exp(n*1j*np.pi/2) * In(z) = (j^n) * In(z)
        - the recurrence relation numerically breaks down for large n and small z due to numerical errors.

    Args:
        n (int): integer order
        z (torch.Tensor): complex argument

    Raises:
        Exception: mixed real, imag argument

    Returns:
        torch.Tensor: result
    """
    z = torch.as_tensor(z, dtype=DTYPE_COMPLEX)

    # pure real arg: use recurrence
    if torch.count_nonzero(z.imag) == 0:
        if n < 0:
            return ((-1) ** n) * Jn(-n, z.real)
        elif n == 0:
            return torch.special.bessel_j0(z.real)
        elif n == 1:
            return torch.special.bessel_j1(z.real)
        else:
            J_nm1 = Jn(n - 1, z.real)
            J_nm2 = Jn(n - 2, z.real)
            return (2 * (n - 1) / z.real) * J_nm1 - J_nm2

    # pure imaginary arg: use modified Bessel functions
    elif torch.count_nonzero(z.real) == 0:
        return (1j**n) * In(n, z.imag)

    # mixed complex args not supported
    else:
        raise Exception(
            "Implementation of Bessel functions supports "
            + "only purely real or purely imaginary argument."
        )


# Bessel functions of second kind
def Yn(n: int, z: torch.Tensor):
    """integer order Bessel functions of second kind via recurrence formula

    Notes:
        - For pure real arguments, use recurrence: Y_n+1 = (2n/z) Y_n - Y_n-1
        - for purely imaginary arguments, use identity via modif. Bessel of first kind
          Yn(1j*z) = exp((n+1)*1j*np.pi/2) * In(z) - 2/np.pi * exp(-1*n*1j*np.pi/2) * Kn(z)
        - the recurrence relation numerically breaks down for large n and small z due to numerical errors.

    Args:
        n (int): integer order
        z (torch.Tensor): complex argument

    Raises:
        Exception: mixed real, imag argument

    Returns:
        torch.Tensor: result
    """
    z = torch.as_tensor(z, dtype=DTYPE_COMPLEX)

    # pure real arg: use recurrence
    if torch.count_nonzero(z.imag) == 0:
        if n < 0:
            return ((-1) ** n) * Yn(-n, z.real)
        elif n == 0:
            return torch.special.bessel_y0(z.real)
        elif n == 1:
            return torch.special.bessel_y1(z.real)
        else:
            Y_nm1 = Yn(n - 1, z.real)
            Y_nm2 = Yn(n - 2, z.real)
            return (2 * (n - 1) / z.real) * Y_nm1 - Y_nm2

    # pure imaginary arg: use modified Bessel functions
    elif torch.count_nonzero(z.real) == 0:
        fact_exp2 = torch.exp(torch.as_tensor((n + 1) * 1j * torch.pi / 2))
        _Yn_t1 = fact_exp2 * In(n, z.imag)

        fact_exp3 = torch.exp(torch.as_tensor(-1 * n * 1j * torch.pi / 2))
        _Yn_t2 = -2 / torch.pi * fact_exp3 * Kn(n, z.imag)

        ynz = _Yn_t1 + _Yn_t2
        return ynz

    # mixed complex args not supported
    else:
        raise Exception(
            "Implementation of Bessel functions supports "
            + "only purely real or purely imaginary argument."
        )


# modified Bessel functions of first kind
def In(n: int, z: torch.Tensor):
    """integer order modified Bessel functions of first kind via recurrence formula

    Notes:
        - use recurrence: I_n+1 = -(2n/z) I_n + I_n-1
        - supports only purely real or purely imaginary arguments
        - the recurrence relation numerically breaks down for large n and small z due to numerical errors.

    Args:
        n (int): integer order
        z (torch.Tensor): complex argument

    Returns:
        torch.Tensor: result
    """
    if n < 0:
        return In(-n, z)
    elif n == 0:
        return torch.special.modified_bessel_i0(z)
    elif n == 1:
        return torch.special.modified_bessel_i1(z)
    else:
        I_nm1 = In(n - 1, z)
        I_nm2 = In(n - 2, z)
        return (-2 * (n - 1) / z) * I_nm1 + I_nm2


# modified Bessel functions of second kind
def Kn(n: int, z: torch.Tensor):
    """integer order modified Bessel functions of second kind via recurrence formula

    Notes:
        - use recurrence: K_n+1 = (2n/z) K_n + K_n-1
        - supports only purely real or purely imaginary arguments
        - the recurrence relation numerically breaks down for large n and small z due to numerical errors.

    Args:
        n (int): integer order
        z (torch.Tensor): complex argument

    Returns:
        torch.Tensor: result
    """
    if n < 0:
        return Kn(-n, z)
    elif n == 0:
        return torch.special.modified_bessel_k0(z)
    elif n == 1:
        return torch.special.modified_bessel_k1(z)
    else:
        K_nm1 = Kn(n - 1, z)
        K_nm2 = Kn(n - 2, z)
        return ((2 * (n - 1)) / z) * K_nm1 + K_nm2


def H1n(n: int, z: torch.Tensor):
    """integer order Hankel functions of first kind

    Notes:
        - use: H1_n = J_n + 1j * Y_n
        - supports only purely real or purely imaginary arguments
        - the recurrence relations used in the Bessel functions numerically break down for large n and small z due to numerical errors.

    Args:
        n (int): integer order
        z (torch.Tensor): complex argument

    Returns:
        torch.Tensor: result
    """
    z = torch.as_tensor(z, dtype=DTYPE_COMPLEX)

    return Jn(n, z) + 1j * Yn(n, z)


def H2n(n: int, z: torch.Tensor):
    """integer order Hankel functions of second kind

    Notes:
        - use: H2_n = J_n - 1j * Y_n
        - supports only purely real or purely imaginary arguments
        - the recurrence relations used in the Bessel functions numerically break down for large n and small z due to numerical errors.

    Args:
        n (int): integer order
        z (torch.Tensor): complex argument

    Returns:
        torch.Tensor: result
    """
    z = torch.as_tensor(z, dtype=DTYPE_COMPLEX)

    return Jn(n, z) - 1j * Yn(n, z)


if __name__ == "__main__":
    from scipy.special import jv, yv, iv, kv, hankel1, hankel2
    import matplotlib.pyplot as plt

    n = -2
    z = torch.linspace(0.1, 1, 500, dtype=torch.float32)

    # compare with scipy
    # bessel
    j_scipy = jv(n, z.numpy())
    j_torch = Jn(n, z)
    y_scipy = yv(n, z.numpy())
    y_torch = Yn(n, z)

    # modified bessel
    i_scipy = iv(n, z.numpy())
    i_torch = In(n, z)
    k_scipy = kv(n, z.numpy())
    k_torch = Kn(n, z)

    # hankel
    h1_scipy = hankel1(n, z.numpy())
    h1_torch = H1n(n, z)
    h2_scipy = hankel2(n, z.numpy())
    h2_torch = H2n(n, z)

    # - plot
    plt.figure(figsize=(10, 4))
    plt.subplot(231)
    plt.plot(z.numpy(), j_scipy, label="scipy")
    plt.plot(z.numpy(), j_torch.numpy(), dashes=[2, 2], label="J - torch")
    plt.legend()

    plt.subplot(232)
    plt.plot(z.numpy(), y_scipy, label="scipy")
    plt.plot(z.numpy(), y_torch.numpy(), dashes=[2, 2], label="Y - torch")
    plt.legend()

    plt.subplot(234)
    plt.plot(z.numpy(), i_scipy, label="scipy")
    plt.plot(z.numpy(), i_torch.numpy(), dashes=[2, 2], label="I - torch")
    plt.legend()

    plt.subplot(235)
    plt.plot(z.numpy(), k_scipy, label="scipy")
    plt.plot(z.numpy(), k_torch.numpy(), dashes=[2, 2], label="K - torch")
    plt.legend()

    plt.subplot(233)
    plt.plot(z.numpy(), h1_scipy, label="scipy")
    plt.plot(z.numpy(), h1_torch.numpy(), dashes=[2, 2], label="H1 - torch")
    plt.legend()

    plt.subplot(236)
    plt.plot(z.numpy(), h2_scipy, label="scipy")
    plt.plot(z.numpy(), h2_torch.numpy(), dashes=[2, 2], label="H2 - torch")

    plt.legend()

    plt.show()



# -*- coding: utf-8 -*-
"""
vectorized pytorch implementations of free space Green's tensors
"""
# %%
import warnings

import numpy as np
import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.tools.misc import get_default_device
from torchgdm.env.base_classes import EnvironmentBase
from torchgdm.tools.misc import tensor_is_diagonal_and_identical, test_positional_input


# %%
# --- vectorized pytorch
def G0_Ep(
    r_probe: torch.Tensor,
    r_source: torch.Tensor,
    wavelength: float,
    eps_env: float,
    epsilon_dist: float = 1e-8,  # to avoid divergence at r_probe=r_source
):
    """3D free space electric field (at r_probe) Green's tensor for an electric dipole (at r_source)

    Args:
        r_probe (torch.Tensor): probe position
        r_source (torch.Tensor): source position
        wavelength (float): in nm
        eps_env (float): environment permittivity
        epsilon_dist (float, optional): small distance to avoid divergence at r_probe=r_source. Defaults to 1e-8.

    Returns:
        torch.Tensor: 3D electric-electric Green's Dyad
    """
    r_probe, r_source = test_positional_input(r_probe, r_source)

    # positional values
    R = r_probe - r_source  # vector difference(s)
    RR = torch.einsum("...i,...j", R, R)  # broadcasted outer product of each vector R
    lR2 = torch.sum(R * R, dim=-1)
    R2diag = torch.eye(3).unsqueeze(0).to(r_probe.device) * lR2.unsqueeze(-1).unsqueeze(
        -1
    )
    lR = torch.sqrt(lR2 + epsilon_dist)

    # wavenumber
    k0 = 2 * torch.pi / wavelength  # vacuum wave number
    k = k0 * torch.sqrt(eps_env)  # k in env. medium
    k_sqrt_eps_lR = k * lR

    # Compute the combined T_EE tensor
    T_combined = (
        -(k**2) * torch.div(RR - R2diag, lR.pow(3).unsqueeze(-1).unsqueeze(-1))  # T1
        - (1j * k)
        * torch.div((3 * RR - R2diag), lR.pow(4).unsqueeze(-1).unsqueeze(-1))  # T2
        + torch.div((3 * RR - R2diag), lR.pow(5).unsqueeze(-1).unsqueeze(-1))  # T3
    )

    # Calculate the phase term and multiply element-wise with T_combined
    sG0 = torch.exp(1j * k_sqrt_eps_lR) / eps_env
    G0 = T_combined * sG0.unsqueeze(-1).unsqueeze(-1)
    return G0


def G0_Hp(
    r_probe: torch.Tensor,
    r_source: torch.Tensor,
    wavelength: float,
    eps_env: float,
    epsilon_dist: float = 1e-8,  # to avoid divergence at r_probe=r_source
) -> torch.Tensor:
    """3D free space magnetic field (at r_probe) Green's tensor for an electric dipole (at r_source)

    Args:
        r_probe (torch.Tensor): probe position
        r_source (torch.Tensor): source position
        wavelength (float): in nm
        eps_env (float): environment permittivity
        epsilon_dist (float, optional): small distance to avoid divergence at r_probe=r_source. Defaults to 1e-8.

    Returns:
        torch.Tensor: 3D electric-electric Green's Dyad
    """
    r_probe, r_source = test_positional_input(r_probe, r_source)

    # positional values
    R = r_probe - r_source  # vector difference(s)
    Dx = R[..., 0]
    Dy = R[..., 1]
    Dz = R[..., 2]
    lR2 = torch.sum(R**2, dim=-1) + epsilon_dist
    lR = torch.sqrt(lR2)

    # wavenumber
    n_env = torch.sqrt(eps_env)
    k0 = 2 * torch.pi / wavelength
    k02n = k0 * k0 * n_env
    k = n_env * k0
    k_sqrt_eps_lR = k * lR

    # compute the combined T_HE tensor
    T_combined = torch.zeros(
        (*R.size()[:-1], 3, 3), dtype=DTYPE_COMPLEX, device=r_probe.device
    )
    T_combined[..., 0, 1] = 1j * k0 * (Dz / (lR**3)) + k02n * (Dz / lR2)
    T_combined[..., 0, 2] = 1j * k0 * (-Dy / (lR**3)) + k02n * (-Dy / lR2)
    T_combined[..., 1, 2] = 1j * k0 * (Dx / (lR**3)) + k02n * (Dx / lR2)
    T_combined[..., 1, 0] = -T_combined[..., 0, 1]
    T_combined[..., 2, 0] = -T_combined[..., 0, 2]
    T_combined[..., 2, 1] = -T_combined[..., 1, 2]

    # Phase term
    sG0 = -1 * torch.exp(1j * k_sqrt_eps_lR)
    T_combined = T_combined * sG0.unsqueeze(-1).unsqueeze(-1)

    return T_combined


def G0_Em(r_probe, r_source, wavelength, eps_env, epsilon_dist: float = 1e-8):
    """3D free space electric field (at r_probe) Green's tensor for a magnetic dipole (at r_source)

    G0_Em = - G0_Hp

    for doc, see :func:`G0_Hp`
    """
    return -1 * G0_Hp(r_probe, r_source, wavelength, eps_env, epsilon_dist=epsilon_dist)


def G0_Hm(r_probe, r_source, wavelength, eps_env, epsilon_dist: float = 1e-8):
    """3D free space magnetic field (at r_probe) Green's tensor for a magnetic dipole (at r_source)

    G0_Hm = G0_Ep

    for doc, see :func:`G0_Ep`
    """
    return G0_Ep(r_probe, r_source, wavelength, eps_env, epsilon_dist=epsilon_dist)


def G0_Ep_ff(
    r_probe: torch.Tensor,
    r_source: torch.Tensor,
    wavelength: float,
    eps_env: float,
    epsilon_dist: float = 1e-8,  # to avoid divergence at r_probe=r_source
):
    """3D far-field approximation for electric-electric Green's tensor

    Args:
        r_probe (torch.Tensor): probe position
        r_source (torch.Tensor): source position
        wavelength (float): in nm
        eps_env (float): environment permittivity
        epsilon_dist (float, optional): small distance to avoid divergence at r_probe=r_source. Defaults to 1e-8.

    Returns:
        torch.Tensor: 3D electric-electric far-field Green's Dyad
    """
    r_probe, r_source = test_positional_input(r_probe, r_source)

    # positional values
    R = r_probe - r_source  # vector difference(s)
    RR = torch.einsum("...i,...j", R, R)  # broadcasted outer product of each vector R
    lR2 = torch.sum(R * R, dim=-1)
    R2diag = torch.eye(3).unsqueeze(0).to(r_probe.device) * lR2.unsqueeze(-1).unsqueeze(
        -1
    )
    lR = torch.sqrt(lR2 + epsilon_dist)

    # wavenumber
    k0 = 2 * torch.pi / wavelength  # vacuum wave number
    k = k0 * torch.sqrt(eps_env + epsilon_dist)  # k in env. medium
    k_sqrt_eps_lR = k * lR

    # Compute the combined T_EE tensor
    T_combined = -(k**2) * torch.div(
        RR - R2diag, lR.pow(3).unsqueeze(-1).unsqueeze(-1)
    )  # T1

    # Calculate the phase term and multiply element-wise with T_combined
    sG0 = torch.exp(1j * k_sqrt_eps_lR) / eps_env
    G0 = T_combined * sG0.unsqueeze(-1).unsqueeze(-1)
    return G0


def G0_Em_ff(
    r_probe: torch.Tensor,
    r_source: torch.Tensor,
    wavelength: float,
    eps_env: float,
    epsilon_dist: float = 1e-8,  # to avoid divergence at r_probe=r_source
) -> torch.Tensor:
    """3D far-field approximation for electric-magnetic Green's tensor

    Args:
        r_probe (torch.Tensor): probe position
        r_source (torch.Tensor): source position
        wavelength (float): in nm
        eps_env (float): environment permittivity
        epsilon_dist (float, optional): small distance to avoid divergence at r_probe=r_source. Defaults to 1e-8.

    Returns:
        torch.Tensor: 3D electric-electric far-field Green's Dyad
    """
    r_probe, r_source = test_positional_input(r_probe, r_source)

    # positional values
    R = r_probe - r_source  # vector difference(s)
    Dx = R[..., 0]
    Dy = R[..., 1]
    Dz = R[..., 2]
    lR2 = torch.sum(R**2, dim=-1) + epsilon_dist
    lR = torch.sqrt(lR2)

    # wavenumber
    k0 = 2 * torch.pi / wavelength
    k = torch.sqrt(eps_env + epsilon_dist) * k0
    k02n = k * k0
    k_sqrt_eps_lR = k * lR

    # compute the combined T_HE tensor
    T_combined = torch.zeros(
        (*R.size()[:-1], 3, 3), dtype=DTYPE_COMPLEX, device=r_probe.device
    )
    T_combined[..., 0, 1] = 1j * k02n * (Dz / lR2)
    T_combined[..., 0, 2] = 1j * k02n * (-Dy / lR2)
    T_combined[..., 1, 2] = 1j * k02n * (Dx / lR2)
    T_combined[..., 1, 0] = -T_combined[..., 0, 1]
    T_combined[..., 2, 0] = -T_combined[..., 0, 2]
    T_combined[..., 2, 1] = -T_combined[..., 1, 2]

    # Phase term
    sG0 = -1 * torch.exp(1j * k_sqrt_eps_lR)
    T_combined = T_combined * sG0.unsqueeze(-1).unsqueeze(-1)

    # Em = -1*Hp:
    return -1 * T_combined


# --- environment class
class EnvHomogeneous3D(EnvironmentBase):
    """class defining a homogeneous environment

    - defines set of free-space Green's tensors
    - environemnt material needs to be isotropic (scalar epsilon) and lossless.
    """

    __name__ = "homogeneous environment 3D"

    def __init__(self, env_material=1.0, device: torch.device = None):
        """class defining a homogeneous 3D environment

        Args:
            env_material (float, optional): Environment material. Either float or class from :mod:`torchgdm.materials`. A float value is interpreted as permittivity.  Defaults to 1.0.
            device (torch.device, optional): Defaults to "cpu".
        """
        super().__init__(device=device)
        self.n_dim = 3

        if type(env_material) in (float, int):
            from torchgdm.materials import MatConstant

            self.env_material = MatConstant(env_material)
        else:
            self.env_material = env_material

    def __repr__(self, verbose=False):
        """description about simulation environment defined by set of dyads"""
        out_str = " ------ homogeneous 3D environment -------"
        out_str += "\n env. material: {}".format(self.env_material)
        # out_str += "\n --------------------------------------"
        return out_str

    def set_device(self, device):
        """move all tensors of the class to device"""
        super().set_device(device)
        self.env_material.set_device(device)

    def get_environment_permittivity_scalar(
        self, wavelength: float, r_probe: torch.Tensor
    ):
        """return scalar complex environment permittivity for `wavelength` at pos. `r_probe` (units in nm)

        assume isotropic env. material: returns [0,0]-component of eps. tensor

        homogeneous medium: all positions `r_probe` will be same material

        Args:
            wavelength (float): in nm
            r_probe (torch.Tensor): position(s) at which the permittivity is to be returned

        Raises:
            ValueError: In case env. material is not isotropic

        Returns:
            torch.Tensor: scalar permittivity at all positions `r_probe`
        """

        eps_env_tensor = self.env_material.get_epsilon(wavelength)

        if not tensor_is_diagonal_and_identical(eps_env_tensor):
            raise ValueError(
                "Environment material is not isotropic, but tensorial epsilon was received. "
                + "Only isotropic environment media are supported."
            )
        self.eps_env_scalar = eps_env_tensor[0, 0]

        if self.eps_env_scalar.imag != 0:
            warnings.warn(
                "Environment permittivity evaluated to non-zero imaginary part. "
                + "This is not physically meaningful in the infinite environment medium."
            )

        eps_env_tensor = (
            torch.ones(len(r_probe), dtype=DTYPE_COMPLEX, device=self.device)
            * self.eps_env_scalar
        )

        return eps_env_tensor

    # --- individual sources Green's tensors
    def get_G_Ep(self, r_probe, r_source, wavelength):
        """3D free space electric field (at r_probe) Green's tensor for an electric dipole (at r_source)

        Args:
            r_probe (torch.Tensor): probe position
            r_source (torch.Tensor): source position
            wavelength (float): in nm

        Returns:
            torch.Tensor: electric-electric Green's Dyad
        """
        eps_env = self.get_environment_permittivity_scalar(
            wavelength,
            r_probe=torch.zeros((1, 3), dtype=DTYPE_FLOAT, device=self.device),
        )[0]
        if r_probe.device != eps_env.device or r_source.device != eps_env.device:
            r_probe.to(self.device)
            r_source.to(self.device)
        return G0_Ep(r_probe, r_source, wavelength, eps_env=eps_env)

    def get_G_Hp(self, r_probe, r_source, wavelength):
        """3D free space magnetic field (at r_probe) Green's tensor for a magnetic dipole (at r_source)

        Args:
            r_probe (torch.Tensor): probe position
            v (torch.Tensor): source position
            wavelength (float): in nm

        Returns:
            torch.Tensor: magnetic-magnetic Green's Dyad
        """
        eps_env = self.get_environment_permittivity_scalar(
            wavelength,
            r_probe=torch.zeros((1, 3), dtype=DTYPE_FLOAT, device=self.device),
        )[0]
        return G0_Hp(r_probe, r_source, wavelength, eps_env=eps_env)

    def get_G_Em(
        self, r_probe: torch.Tensor, r_source: torch.Tensor, wavelength: float
    ):
        """3D free space electric field (at r_probe) Green's tensor for a magnetic dipole (at r_source)

        G0_Em = - G0_Hp

        for doc, see :meth:`EnvHomogeneous3D.get_G_Hp`
        """
        eps_env = self.get_environment_permittivity_scalar(
            wavelength,
            r_probe=torch.zeros((1, 3), dtype=DTYPE_FLOAT, device=self.device),
        )[0]
        return G0_Em(r_probe, r_source, wavelength, eps_env=eps_env)

    def get_G_Hm(
        self, r_probe: torch.Tensor, r_source: torch.Tensor, wavelength: float
    ):
        """3D free space magnetic field (at r_probe) Green's tensor for a magnetic dipole (at r_source)

        G0_Hm = G0_Ep

        for doc, see :meth:`EnvHomogeneous3D.get_G_Ep`
        """
        eps_env = self.get_environment_permittivity_scalar(
            wavelength,
            r_probe=torch.zeros((1, 3), dtype=DTYPE_FLOAT, device=self.device),
        )[0]
        return G0_Hm(r_probe, r_source, wavelength, eps_env=eps_env)

    # --- far-field approximation
    def get_G_Ep_farfield(self, r_probe, r_source, wavelength):
        """Electric field asymptotic far-field Green's tensor for an electric dipole

        r_probe: field evaluation position, r_source: emitter position
        """
        eps_env = self.get_environment_permittivity_scalar(
            wavelength,
            r_probe=torch.zeros((1, 3), dtype=DTYPE_FLOAT, device=self.device),
        )[0]
        return G0_Ep_ff(r_probe, r_source, wavelength, eps_env=eps_env)

    def get_G_Em_farfield(self, r_probe, r_source, wavelength):
        """Electric field asymptotic far-field Green's tensor for a magnetic dipole

        r_probe: field evaluation position, r_source: emitter position
        """
        eps_env = self.get_environment_permittivity_scalar(
            wavelength,
            r_probe=torch.zeros((1, 3), dtype=DTYPE_FLOAT, device=self.device),
        )[0]
        return G0_Em_ff(r_probe, r_source, wavelength, eps_env=eps_env)


## ---------------- test Green's tensors ----------------
if __name__ == "__main__":
    import time

    import scipy
    import numpy as np
    import matplotlib.pyplot as plt

    from torchgdm.env.freespace_3d._dyads_numpy import (
        G0_Ep_vec,
        G0_Hp_vec,
        G0_Em_vec,
        G0_Hm_vec,
    )
    from pyGDM2 import tools
    from pyGDM2.propagators import _G0 as G0_Ep_seq
    from pyGDM2.propagators import _G0_HE as G0_Hp_seq

    device = torch.device("cuda:0")
    # torch.cuda.empty_cache()
    # device = torch.device('cpu')

    func_seq = G0_Ep_seq
    func_np = G0_Ep_vec
    func_torch = G0_Ep

    func_seq = G0_Hp_seq
    func_np = G0_Hp_vec
    func_torch = G0_Hp

    # func_np = G0_Em_vec
    # func_torch = G0_Em
    # func_np = G0_Hm_vec
    # func_torch = G0_Hm

    # ---- technical test
    N_dim1 = 10
    R1 = torch.randn(N_dim1, 3).to(device)
    R2 = torch.randn(N_dim1, 3).to(device)
    wavelength = 500.0  # nm
    eps = 2.25  # permittivity
    result = func_torch(R1, R2, wavelength, eps)
    print(result.shape)

    # ---- quantitative test
    # setup
    eps_env = 1.5
    wl = 500.0

    DX, NX = 10, 20
    Z0 = 10
    r_probe = tools.generate_NF_map_XY(-DX, DX, NX, -DX, DX, NX, Z0)
    r0 = np.zeros_like(r_probe, dtype=np.float32)  # fix emitter position
    r_probe_t = torch.as_tensor(r_probe, dtype=DTYPE_FLOAT).to(device)
    r0_t = torch.as_tensor(r0, dtype=DTYPE_FLOAT).to(device)

    # eval numpy
    print("evaluating {} greens tensors".format(len(r_probe)))
    t1a = time.time()
    G0_vec1 = func_np(r0, r_probe, wl, eps_env)
    t1b = time.time()
    print("time numpy: {:.5f}s".format(t1b - t1a))

    # eval sequential
    # print('evaluating {} greens tensors'.format(len(r_probe)))
    # t1a = time.time()
    # for i in range(len(r0)):
    #     xx, yy, zz, xy, xz, yx, yz, zx, zy = func_seq(r0[i], r_probe[i], wl, eps_env)
    #     G0 = np.array([
    #         [xx, xy, xz],
    #         [yx, yy, yz],
    #         [zx, zy, zz],
    #         ]).astype(np.complex64)
    #     G0_vec1[i] = G0
    # t1b = time.time()
    # print('time sequential: {:.5f}s'.format(t1b-t1a))

    # eval torch
    t2a = time.time()
    G0_vec2t = func_torch(r0_t, r_probe_t, wl, eps_env)
    t2b = time.time()

    # evaluate using broadcasting
    # interact_NxN = func_torch(r_probe_t.unsqueeze(1), r_probe_t.unsqueeze(0), wl, eps_env)

    print("time torch (on device {}): {:.5f}s".format(device, t2b - t2a))
    print("speed-up torch: x{:.3f}".format((t1b - t1a) / (t2b - t2a)))

    G0_vec2 = G0_vec2t.cpu().numpy()
    G0_vec1 = G0_vec1.reshape(NX, NX, 3, 3)
    G0_vec2 = G0_vec2.reshape(NX, NX, 3, 3)

    print(np.abs(G0_vec1).mean())
    print((np.abs(G0_vec1) - np.abs(G0_vec2)).mean())

    i1, i2 = 2, 0
    plt.subplot(221)
    plt.title("numpy")
    plt.imshow(G0_vec1[..., i1, i2].real)
    plt.colorbar()
    plt.subplot(222)
    plt.title("pytorch")
    plt.imshow(G0_vec2[..., i1, i2].real)
    plt.colorbar()
    plt.subplot(223)
    plt.imshow(G0_vec1[..., i1, i2].imag)
    plt.colorbar()
    plt.subplot(224)
    plt.imshow(G0_vec2[..., i1, i2].imag)
    plt.colorbar()

    plt.show()

    # ---- test inversion speed
    print("test inversion: {}x{} dipoles".format(NX, NX))

    t3a = time.time()
    G0_vec1 = G0_vec1.reshape(3 * NX, 3 * NX)
    inv = scipy.linalg.lu(G0_vec1)
    t3b = time.time()
    print("time scipy-LU (on CPU): {:.5f}s".format(t3b - t3a))

    t4a = time.time()
    G0_vec2t = G0_vec2t.reshape(3 * NX, 3 * NX)
    inv = torch.linalg.lu(G0_vec2t)
    t4b = time.time()
    print("time torch-LU (on device {}): {:.5f}s".format(device, t4b - t4a))

    # %%
    # t0 = time.time()
    # i1 = torch.linalg.inv(G0_vec2t)
    # t1 = time.time()
    # i2 = torch.linalg.inv_ex(G0_vec2t)
    # t2 = time.time()
    # print(t1-t0)
    # print(t2-t1)



# encoding=utf-8
# %%
import unittest

import torch

import torchgdm as tg


class TestGlobalSizeScaling(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print("testing global size scaling invariance...")

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

        self.scale_factors = [1.0, 1.5]

        # --- setup a test case
        # - materials
        self.mat_struct = tg.materials.MatConstant(eps=16)
        self.mat_env = tg.materials.MatConstant(eps=1.33)
        # - environment
        self.env = tg.env.freespace_3d.EnvHomogeneous3D(env_material=self.mat_env)

        # - illumination field(s)
        self.plane_wave = tg.env.freespace_3d.PlaneWave(e0p=0.7, e0s=-0.5)

    def test_calculate_E_H(self):
        for device in self.devices:
            Q_sc = []
            Q_ex = []
            for scale in self.scale_factors:

                # scale wavelengths
                wavelengths = torch.linspace(550.0, 750.0, 3) * scale

                # scale geometry
                step = 20 * scale
                structure = tg.struct.StructDiscretizedCubic3D(
                    tg.struct.volume.cube(l=150 * scale / step), step, self.mat_struct
                )
                cs_geo = tg.tools.geometry.get_geometric_crosssection(structure)

                # - define and run simulation
                sim = tg.Simulation(
                    structures=[structure],
                    illumination_fields=[self.plane_wave],
                    environment=self.env,
                    wavelengths=wavelengths,
                    device=device,
                )
                sim.run(verbose=False, progress_bar=False)
                cs_results = tg.tools.batch.calc_spectrum(
                    sim, tg.postproc.crosssect.total, progress_bar=False
                )
                Q_sc.append(cs_results["scs"] / cs_geo)
                Q_ex.append(cs_results["ecs"] / cs_geo)

            torch.testing.assert_close(Q_sc[0], Q_sc[1], rtol=1e-3, atol=0.01)
            torch.testing.assert_close(Q_ex[0], Q_ex[1], rtol=1e-3, atol=0.01)
            if self.verbose:
                print("  - {}: scaling test passed.".format(device))


class TestGlobalWavelengthAndMaterialScaling(unittest.TestCase):

    def setUp(self):
        self.verbose = False
        if self.verbose:
            print("testing global wavelength / material scaling invariance...")

        # --- determine if GPU is available
        self.devices = ["cpu"]
        if torch.cuda.is_available():
            self.devices.append("cuda:0")

        # test case
        # test case setup
        self.r_sphere = 80.0  # nm
        self.step = 25.0

        self.n_environment = 1.3
        self.n_sphere = 4.0

        # set up environments and structures with non-vacuum environment and
        # an equivalent case which instead uses the scaling n=n/n_env and
        # wavelength=wavelength/n_env
        self.n_env1 = self.n_environment
        self.n_sph1 = self.n_sphere
        self.wls1 = torch.linspace(600.0, 800.0, 3)
        self.r_probe = tg.tools.geometry.coordinate_map_2d_square(
            500, 15, r3=0, delta2=750
        )

        self.n_env2 = 1.0
        self.n_sph2 = self.n_sphere / self.n_environment
        self.wls2 = self.wls1 / self.n_environment

        self.mat_env1 = tg.materials.MatConstant(self.n_env1**2)
        self.mat_env2 = tg.materials.MatConstant(self.n_env2**2)
        self.env1 = tg.env.freespace_3d.EnvHomogeneous3D(env_material=self.mat_env1)
        self.env2 = tg.env.freespace_3d.EnvHomogeneous3D(env_material=self.mat_env2)

        # - illumination field
        self.e_inc = tg.env.freespace_3d.PlaneWave(e0p=1, e0s=0)

        # - test-structure: sphere
        self.mat1 = tg.materials.MatConstant(eps=self.n_sph1**2)
        self.s_mesh1 = tg.struct.volume.StructDiscretizedCubic3D(
            tg.struct.volume.geometries.sphere(self.r_sphere / self.step),
            self.step,
            self.mat1,
        )
        self.s_mesh1.set_center_of_mass([0, 0, 0])

        self.mat2 = tg.materials.MatConstant(eps=self.n_sph2**2)
        self.s_mesh2 = tg.struct.volume.StructDiscretizedCubic3D(
            tg.struct.volume.geometries.sphere(self.r_sphere / self.step),
            self.step,
            self.mat2,
        )
        self.s_mesh2.set_center_of_mass([0, 0, 0])

    def test_full_sim(self):
        for device in self.devices:
            # test polarizabilities
            self.s_mesh1.set_device(device)
            self.s_mesh2.set_device(device)
            for i_wl in range(len(self.wls1)):
                a_m1 = self.s_mesh1.get_polarizability_6x6(self.wls1[i_wl], self.env1)
                a_m2 = self.s_mesh2.get_polarizability_6x6(self.wls2[i_wl], self.env2)
                torch.testing.assert_close(
                    ((a_m1[0, 0, 0] / a_m2[0, 0, 0]) ** 0.5).real,
                    torch.as_tensor(self.n_environment, device=a_m1.device),
                )
                torch.testing.assert_close(
                    ((a_m1[0, 0, 0] / a_m2[0, 0, 0]) ** 0.5).imag,
                    torch.as_tensor(0.0, device=a_m1.device),
                )

            # test sim. results
            sim_f1 = tg.Simulation(
                structures=[self.s_mesh1],
                environment=self.env1,
                illumination_fields=self.e_inc,
                wavelengths=self.wls1,
            )
            sim_f1.run(verbose=False, progress_bar=False)
            sim_f2 = tg.Simulation(
                structures=[self.s_mesh2],
                environment=self.env2,
                illumination_fields=self.e_inc,
                wavelengths=self.wls2,
            )
            sim_f2.run(verbose=False, progress_bar=False)

            # test cross sections
            cs_1 = sim_f1.get_crosssections(progress_bar=False)
            cs_2 = sim_f2.get_crosssections(progress_bar=False)

            torch.testing.assert_close(cs_1["ecs"], cs_2["ecs"], atol=10, rtol=1e-5)
            torch.testing.assert_close(cs_1["acs"], cs_2["acs"], atol=10, rtol=1e-5)
            torch.testing.assert_close(cs_1["scs"], cs_2["scs"], atol=10, rtol=1e-5)

            # test fields
            for i_wl in range(len(self.wls1)):
                nf1 = sim_f1.get_nearfield(
                    self.wls1[i_wl], r_probe=self.r_probe, progress_bar=False
                )
                nf2 = sim_f2.get_nearfield(
                    self.wls2[i_wl], r_probe=self.r_probe, progress_bar=False
                )

                for which in ["sca", "tot", "inc"]:
                    torch.testing.assert_close(
                        nf1[which].get_efield(),
                        nf2[which].get_efield(),
                        atol=0.001,
                        rtol=1e-4,
                    )
                    torch.testing.assert_close(
                        nf1[which].get_hfield(),
                        nf2[which].get_hfield() * self.n_environment,
                        atol=0.001,
                        rtol=1e-4,
                    )

            if self.verbose:
                print(
                    "  - {}: full sim environment index scaling test passed.".format(
                        device
                    )
                )

    def test_eff_pola(self):
        for device in self.devices:
            # eff. pola model
            self.s_mesh1.set_device(device)
            self.s_mesh2.set_device(device)
            self.s_eff1 = self.s_mesh1.convert_to_effective_polarizability_pair(
                environment=self.env1,
                wavelengths=self.wls1,
                progress_bar=False,
                verbose=False,
            )
            self.s_eff2 = self.s_mesh2.convert_to_effective_polarizability_pair(
                environment=self.env2,
                wavelengths=self.wls2,
                progress_bar=False,
                verbose=False,
            )

            # test polarizabilities
            torch.testing.assert_close(
                self.s_eff1.alpha_data[..., 0, 0],
                self.s_eff2.alpha_data[..., 0, 0] * self.n_environment**2,
                atol=10,
                rtol=1e-5,
            )
            torch.testing.assert_close(
                self.s_eff1.alpha_data[..., 3, 3],
                self.s_eff2.alpha_data[..., 3, 3] * self.n_environment,
                atol=10,
                rtol=1e-5,
            )

            # test sim. results
            sim_eff1 = tg.Simulation(
                structures=[self.s_eff1],
                environment=self.env1,
                illumination_fields=self.e_inc,
                wavelengths=self.wls1,
            )
            sim_eff1.run(verbose=False, progress_bar=False)
            sim_eff2 = tg.Simulation(
                structures=[self.s_eff2],
                environment=self.env2,
                illumination_fields=self.e_inc,
                wavelengths=self.wls2,
            )
            sim_eff2.run(verbose=False, progress_bar=False)

            # test cross sections
            cs_a1 = sim_eff1.get_crosssections(progress_bar=False)
            cs_a2 = sim_eff2.get_crosssections(progress_bar=False)

            torch.testing.assert_close(cs_a1["ecs"], cs_a2["ecs"], atol=10, rtol=1e-5)
            torch.testing.assert_close(cs_a1["scs"], cs_a2["scs"], atol=1000, rtol=1e-2)

            # test fields
            for i_wl in range(len(self.wls1)):
                nf_eff1 = sim_eff1.get_nearfield(
                    self.wls1[i_wl], r_probe=self.r_probe, progress_bar=False
                )
                nf_eff2 = sim_eff2.get_nearfield(
                    self.wls2[i_wl], r_probe=self.r_probe, progress_bar=False
                )

                for which in ["sca", "tot", "inc"]:
                    torch.testing.assert_close(
                        nf_eff1[which].get_efield(),
                        nf_eff2[which].get_efield(),
                        atol=0.1,
                        rtol=1e-1,
                    )
                    torch.testing.assert_close(
                        nf_eff1[which].get_hfield(),
                        nf_eff2[which].get_hfield() * self.n_environment,
                        atol=0.1,
                        rtol=1e-1,
                    )

            if self.verbose:
                print(
                    "  - {}: effective polarizabilities env. index scaling test passed.".format(
                        device
                    )
                )

    def test_mie_pola(self):
        for device in self.devices:
            # Mie eff. pola
            self.r_sphere_mie = self.r_sphere

            self.s_mie1 = tg.struct.StructMieSphereEffPola3D(
                wavelengths=self.wls1,
                environment=self.env1,
                radii=self.r_sphere_mie,
                materials=self.mat1,
                device=device,
            )

            self.s_mie2 = tg.struct.StructMieSphereEffPola3D(
                wavelengths=self.wls2,
                environment=self.env2,
                radii=self.r_sphere_mie,
                materials=self.mat2,
                device=device,
            )

            # test polarizabilities
            torch.testing.assert_close(
                self.s_mie1.alpha_data[..., 0, 0],
                self.s_mie2.alpha_data[..., 0, 0] * self.n_environment**2,
                atol=10,
                rtol=1e-5,
            )
            torch.testing.assert_close(
                self.s_mie1.alpha_data[..., 3, 3],
                self.s_mie2.alpha_data[..., 3, 3] * self.n_environment,
                atol=10,
                rtol=1e-5,
            )

            # test sim. results
            sim_mie1 = tg.Simulation(
                structures=[self.s_mie1],
                environment=self.env1,
                illumination_fields=self.e_inc,
                wavelengths=self.wls1,
            )
            sim_mie1.run(verbose=False, progress_bar=False)
            sim_mie2 = tg.Simulation(
                structures=[self.s_mie2],
                environment=self.env2,
                illumination_fields=self.e_inc,
                wavelengths=self.wls2,
            )
            sim_mie2.run(verbose=False, progress_bar=False)

            # test cross sections
            cs_mie1 = sim_mie1.get_crosssections(progress_bar=False)
            cs_mie2 = sim_mie2.get_crosssections(progress_bar=False)

            torch.testing.assert_close(
                cs_mie1["ecs"], cs_mie2["ecs"], atol=10, rtol=1e-5
            )
            torch.testing.assert_close(
                cs_mie1["scs"], cs_mie2["scs"], atol=1000, rtol=1e-2
            )

            # test fields
            for i_wl in range(len(self.wls1)):
                nf_mie1 = sim_mie1.get_nearfield(
                    self.wls1[i_wl], r_probe=self.r_probe, progress_bar=False
                )
                nf_mie2 = sim_mie2.get_nearfield(
                    self.wls2[i_wl], r_probe=self.r_probe, progress_bar=False
                )

                for which in ["sca", "tot", "inc"]:
                    torch.testing.assert_close(
                        nf_mie1[which].get_efield(),
                        nf_mie2[which].get_efield(),
                        atol=0.1,
                        rtol=1e-1,
                    )
                    torch.testing.assert_close(
                        nf_mie1[which].get_hfield(),
                        nf_mie2[which].get_hfield() * self.n_environment,
                        atol=0.1,
                        rtol=1e-1,
                    )

            if self.verbose:
                print(
                    "  - {}: mie effective polarizabilities env. index scaling test passed.".format(
                        device
                    )
                )


# %%
if __name__ == "__main__":
    print("testing global scaling invariance...")
    torch.set_printoptions(precision=7)
    unittest.main(argv=["first-arg-is-ignored"], exit=False)



# -*- coding: utf-8 -*-
"""volume discretization: geometry-definitions and discretizer
"""
import warnings

import torch

from torchgdm.constants import DTYPE_FLOAT, DTYPE_COMPLEX
from torchgdm.tools.misc import get_default_device


# ==============================================================================
# geometry definitions
# ==============================================================================
def cuboid(l: int, w: int, h: int):
    """cuboid with dimensions `l` along x, `w` along y, `h` along z

    Args:
        l (int): length (units of step)
        w (int): width (units of step)
        h (int): height (units of step)

    Returns:
        func: condition(x,y,z): whether to put or not a mesh-cell at coordinate
        tuple: discretizer walk limits necessary to contain full structure
    """

    def condition(xi, yi, zi):
        in_cuboid = (
            -l / 2.0 < xi <= l / 2.0
            and -w / 2.0 < yi <= w / 2.0
            and 0 <= zi < h - (1 / 2) ** 0.5
        )
        return in_cuboid

    mesh_limits = (-int(l), int(l) + 1, -int(w), int(w) + 1, -1, 2 * int(h) + 1)

    return condition, mesh_limits


def cube(l: int):
    """cube with side length `l`

    Args:
        l (int): side length (units of step)

    Returns:
        func: condition(x,y,z): whether to put or not a mesh-cell at coordinate
        tuple: discretizer walk limits necessary to contain full structure
    """
    return cuboid(l, l, l)


def spheroid(r_x: float, r_y: float, r_z: float):
    """spheroid with radii r_x, r_y, r_z along X, Y, Z axis

    - if r_z > r_y = r_x --> prolate spheroid
    - if r_z < r_y = r_x --> oblate spheroid
    - if r_z = r_y = r_x --> sphere

    contributed by C. Majorel.

    Args:
        r_x (float): X semi-axes length (units of step)
        r_y (float): Y semi-axes length (units of step)
        r_z (float): Z semi-axes length (units of step)

    Returns:
        func: condition(x,y,z): whether to put or not a mesh-cell at coordinate
        tuple: discretizer walk limits necessary to contain full structure
    """

    def condition(xi, yi, zi):
        return (
            ((xi / (r_x + 0.2)) ** 2)
            + ((yi / (r_y + 0.2)) ** 2)
            + (zi / (r_z + 0.2)) ** 2
        ) <= 1

    mesh_limits = (
        -2 * int(r_x),
        2 * int(r_x) + 1,
        -2 * int(r_y),
        2 * int(r_y) + 1,
        -2 * int(r_z),
        2 * int(r_z) + 1,
    )

    return condition, mesh_limits


def sphere(r: float):
    """sphere of radius `r` (units of step)"""
    return spheroid(r, r, r)


def ellipse(r1: float, r2: float, h: int):
    """elliptical cylinder with radii `r1`, `r2` (x, y) and height `h` (z)

    Args:
        r1 (float): semi-axis length along `x` (units of step)
        r2 (float): semi-axis length along `y` (units of step)
        h (int):  height along `z` (units of step)

    Returns:
        func: condition(x,y,z): whether to put or not a mesh-cell at coordinate
        tuple: discretizer walk limits necessary to contain full structure
    """

    def condition(xi, yi, zi):
        in_disc = (
            (xi / (r1 + 0.1)) ** 2 + (yi / (r2 + 0.1)) ** 2 <= 1
        ) and 0 <= zi < h - (1 / 2) ** 0.5
        return in_disc

    mesh_limits = (
        -int(r1) - 2,
        int(r1) + 2,
        -int(r2) - 2,
        int(r2) + 2,
        -1,
        2 * int(h) + 1,
    )

    return condition, mesh_limits


def disc(r: float, h: int):
    """cylinder with radius `r` in x and y, and height `h` along z

    Args:
        r (float): radius (units of step)
        h (int): height along z (units of step)

    Returns:
        func: condition(x,y,z): whether to put or not a mesh-cell at coordinate
        tuple: discretizer walk limits necessary to contain full structure
    """
    return ellipse(r, r, h)


def split_ring(r_out: float, r_in: float, h: int, alpha_g: float = 0):
    """split ring

    Args:
        r_out (float): outer radius (units of step)
        r_in (float): inner radius (units of step)
        h (int): height (units of step)
        alpha_g (float, optional): split gap angle (units of rad). Defaults to 0.

    Returns:
        func: condition(x,y,z): whether to put or not a mesh-cell at coordinate
        tuple: discretizer walk limits necessary to contain full structure
    """

    def condition(xi, yi, zi):
        C1 = r_out**2 >= (xi**2 + yi**2) >= r_in**2
        if yi > 0 and alpha_g != 0:
            C2 = not (
                torch.arctan2(torch.as_tensor(abs(xi)), torch.as_tensor(abs(yi)))
                <= alpha_g / 2.0
            )
        else:
            C2 = True

        C3 = 0 <= zi < h - 1 / 2**0.5

        return (C1 and C2) and C3

    mesh_limits = (
        -int(r_out) - 2,
        int(r_out) + 2,
        -int(r_out) - 2,
        int(r_out) + 2,
        -1,
        1.5 * int(h),
    )

    return condition, mesh_limits


def prism_trigonal(d: int, h: int, truncate_edge: float = 0.0):
    """uniform right triangular prism with side length d and height h

    Args:
        d (int): side length  (units of step)
        h (int): height (units of step)
        truncate_edge (float, optional): truncation ratio for the top triangle edge. Defaults to 0.0.

    Returns:
        func: condition(x,y,z): whether to put or not a mesh-cell at coordinate
        tuple: discretizer walk limits necessary to contain full structure
    """
    sqrt3 = 3**0.5
    n_l = d / 2.0
    h_s = truncate_edge * n_l * sqrt3 / 2.0
    h_abs = (sqrt3 / 2.0) * n_l

    def condition(xi, yi, zi):
        in_prism = (
            (abs(xi) <= (h_abs - (yi + 0.5)) / sqrt3)
            and (
                abs(xi)
                <= (((3 * h_abs - sqrt3 * n_l * truncate_edge) + (yi + 0.5)) / sqrt3)
            )
            and (yi <= h_abs - h_s)
            and (yi >= -h_abs)
            and (0 <= zi < h - 1 / (2**0.5))
        )

        return in_prism

    mesh_limits = (
        -1 * int(n_l) - 5,
        int(n_l) + 5,
        -1 * int(n_l) - 5,
        int(n_l) + 5,
        -1 * int(h) - 5,
        int(h) + 5,
    )

    return condition, mesh_limits


def from_image(
    img_file_name,
    pixels_per_step=1,
    h=1,
    threshold=100,
    positive_contrast=True,
    return_image=False,
    center_structure=True,
):
    """planar structure from image based on contrast

    Args:
        img_file_name (string): path to image-file, or numpy array containing image. If numpy array, values should range from 0 to 255.
        pixels_per_step (int, optional): number of pixels in the image corresponding to one discretization step. This is essentially a resize factor. Defaults to 1.
        h (int, optional): height of the planar structure (units of stepsize). Defaults to 1.
        threshold (int, optional): threshold value between [0, 255] to declare a pixel as *structure*. all darker pixels will be interpreted as structure. Defaults to 100.
        positive_contrast (bool, optional): whether to interpret dark (deault) or bright pixels as structure material. Defaults to True.
        return_image (bool, optional): if True, returns a 2D numpy array corresponding to the image AND the structure. Defaults to False.
        center_structure (bool, optional): whether to automatically center structure. Defaults to True.

    Returns:
        if return_image==True:
            numpy ndarray: image pixel values

        if return_image==False, returns:
            func: condition(x,y,z): whether to put or not a mesh-cell at coordinate
            tuple: discretizer walk limits necessary to contain full structure

    """
    from PIL import Image
    import numpy as np

    # - load image
    if type(img_file_name) == str:
        img = Image.open(img_file_name)
        img.load()
    else:
        img = Image.fromarray(img_file_name)

    # - remove alpha channel, if exists
    if len(img.split()) == 4:
        background = Image.new("RGB", img.size, (255, 255, 255))
        background.paste(img, mask=img.split()[3])  # 3 is the alpha channel
    else:
        # RGB data --> no alpha channel
        background = img

    # - rescale image to N pixels per step
    X0 = background.size[0]
    basewidth = int(round(X0 * pixels_per_step))
    hsize = int(
        (float(background.size[1]) * float((basewidth / float(background.size[0]))))
    )
    img = background.resize((basewidth, hsize), Image.LANCZOS)
    img = img.convert("L")

    data = np.asarray(img.getdata()).reshape(img.size[::-1])
    data = np.rot90(data, 3)  # adapt to torchgdm convention: rotate by 270 degrees

    # - mesh size limits
    if center_structure:
        x_min = -(data.shape[0] // 2)
        y_min = -(data.shape[1] // 2)
        x_max = (data.shape[0] + 1) // 2
        y_max = (data.shape[1] + 1) // 2
    else:
        x_min = 0
        y_min = 0
        x_max = data.shape[0]
        y_max = data.shape[1]
    mesh_limits = (x_min - 1, x_max + 1, y_min - 1, y_max + 1, 0, h)

    # - generate structure condition for pixels above threshold
    if positive_contrast:
        data[data <= threshold] = threshold
        data[data > threshold] = 0
    else:
        data[data >= threshold] = threshold
        data[data < threshold] = 0

    def condition(xi, yi, zi):
        x_idx = int(xi - x_min)
        y_idx = int(yi - y_min)
        if x_idx < 0 or x_idx >= data.shape[0]:
            return False  # x outside data range
        elif y_idx < 0 or y_idx >= data.shape[1]:
            return False  # y outside data range
        else:
            return data[x_idx][y_idx] != 0

    if return_image:
        # original image to return
        img_orig = background.convert("L")
        data_img = np.asarray(img_orig.getdata()).reshape(img_orig.size[::-1])
        return data_img
    else:
        print(mesh_limits)
        return condition, mesh_limits


# ==============================================================================
# discretizer
# ==============================================================================
def discretizer_cubic(
    condition_func, limits: tuple, z_offset: float = 0.5, step: float = 1.0
):
    """discretize volume of a 3D structure on a cubic grid

    discretize by evaluating a boolean condition function at each
    position on a 3D grid within given spatial limits

    Args:
        condition_func (func): functional, returning the condition as function of coordinate indices. 'condition_func(xi,yi,zi)', returns a bool. (True: put a meshpoint, False: no meshpoint)
        limits (tuple): discretizer walk range. tuple of 6  ints (x0, x1, y0, y1, z0, z1)
        z_offset (float, optional): additional z-offset, to avoid placing dipoles at z=0 (units of step). Defaults to 0.5.
        step (float, optional): nominal stepsize in nm. Defaults to 1.0.

    Returns:
        torch.Tensor: 3d coordinates of discretization. shape (N, 3) for `N` meshpoints.
    """
    sp = []
    for xi in range(int(limits[0]), int(limits[1])):
        for yi in range(int(limits[2]), int(limits[3])):
            for zi in range(int(limits[4]), int(limits[5])):
                if condition_func(xi, yi, zi):
                    sp.append([xi, yi, zi + z_offset])

    return torch.as_tensor(sp, dtype=DTYPE_FLOAT) * step


def discretizer_hexagonalcompact(
    condition_func,
    limits,
    z_offset=0.5,
    step=1.0,
    orientation=1,
    auto_expand_range=True,
):
    """discretize volume of a 3D structure on a hexagonal compact grid

    discretize by evaluating a boolean condition function at each
    position on a 3D grid within given spatial limits

    Args:
        condition_func (func): functional, returning the condition as function of coordinate indices. 'condition_func(xi,yi,zi)', returns a bool. (True: put a meshpoint, False: no meshpoint)
        limits (tuple): discretizer walk range. tuple of 6  ints (x0, x1, y0, y1, z0, z1)
        z_offset (float, optional): additional z-offset, to avoid placing dipoles at z=0 (units of step). Defaults to 0.5.
        step (float, optional): nominal stepsize in nm. Defaults to 1.0.
        auto_expand_range (bool, optional): expand the discretization volume to account for possible non-integer position values. Defaults to True.

    Returns:
        torch.Tensor: 3d coordinates of discretization. shape (N, 3) for `N` meshpoints.
    """
    sp = []
    x_lim = list(limits[0:2])
    y_lim = list(limits[2:4])
    z_lim = list(limits[4:6])

    # expand discretization volume wrt hexagonal unit cell volume
    # useful e.g. for non-integer position values
    if auto_expand_range:
        vol_scale = 2 / 3**0.5 - 1
        x_lim[0] = x_lim[0] - vol_scale
        x_lim[1] = x_lim[1] + vol_scale

        y_lim[0] = y_lim[0] - vol_scale
        y_lim[1] = y_lim[1] + vol_scale

        z_lim[0] = z_lim[0] - vol_scale
        z_lim[1] = z_lim[1] + vol_scale

    # discretizer orientation 1
    if orientation == 1:
        for zi in range(int(z_lim[0]), int(z_lim[1])):
            z = zi * (2.0 / 3.0) ** 0.5
            for yi in range(int(y_lim[0]), int(y_lim[1])):
                y = yi * 3**0.5 / 2.0 + 2.0 * abs(zi / 2.0 - int(zi / 2.0)) / 3**0.5
                for xi in range(int(x_lim[0]), int(x_lim[1])):
                    x = xi + abs(yi / 2.0 - int(yi / 2.0))
                    if condition_func(x, y, z):
                        sp.append([x, y, z + z_offset])

    # discretizer orientation 2
    else:
        for xi in range(int(x_lim[0]), int(x_lim[1])):
            x = xi * (2.0 / 3.0) ** 0.5
            for yi in range(int(y_lim[0]), int(y_lim[1])):
                y = yi * 3**0.5 / 2.0 + 2.0 * abs(xi / 2.0 - int(xi / 2.0)) / 3**0.5
                for zi in range(int(z_lim[0]), int(z_lim[1])):
                    z = zi + abs(yi / 2.0 - int(yi / 2.0))
                    if condition_func(x, y, z):
                        sp.append([x, y, z + z_offset])

    return torch.as_tensor(sp, dtype=DTYPE_FLOAT) * step


# ==============================================================================
# Transformations - TODO
# ==============================================================================


# %%
if __name__ == "__main__":

    step = 20.0

    # geometry_pos = discretizer_cubic(*spheroid(9,4,4), step=step)
    geometry_pos = discretizer_cubic(*split_ring(10, 7, 2, torch.pi / 4), step=step)

    import matplotlib.pyplot as plt

    plt.scatter(geometry_pos.T[0], geometry_pos.T[1])



# encoding=utf-8
"""
3D visualization tools for geometries
"""
# %%
import copy
import warnings
import itertools

import torch

from torchgdm.constants import COLORS_DEFAULT
from torchgdm.tools.misc import to_np
from torchgdm import tools


# global color handling
COLOR_ITERATOR = itertools.cycle(COLORS_DEFAULT)
LEGEND_ENTRIES_LOOKUP = dict()


# reset color iterator and materials lookup
def _reset_color_iterator():
    global COLOR_ITERATOR, LEGEND_ENTRIES_LOOKUP
    COLOR_ITERATOR = itertools.cycle(COLORS_DEFAULT)
    LEGEND_ENTRIES_LOOKUP = dict()


def _return_next_color():
    return next(COLOR_ITERATOR)


_reset_color_iterator()


def _generate_legend(pl):
    global LEGEND_ENTRIES_LOOKUP

    if len(LEGEND_ENTRIES_LOOKUP) != 0:
        labels = LEGEND_ENTRIES_LOOKUP.keys()
    else:
        labels = []

    pv_labels = []
    for i_s, label in enumerate(labels):
        legend_dict = LEGEND_ENTRIES_LOOKUP[label]
        pv_labels.append([label, legend_dict["fc"]])

    pl.add_legend(labels=pv_labels, bcolor="w", face=None)

    return pl.legend


def _plot_structure_discretized(
    struct,
    scale=1.0,
    color="auto",
    show_grid=True,
    legend=True,
    alpha=1.0,
    show="auto",
    pl=None,
    reset_color_cycle=True,
    **kwargs,
):
    import numpy as np
    import pyvista as pv

    if "projection" in kwargs:
        kwargs.pop("projection")

    global COLOR_ITERATOR, LEGEND_ENTRIES_LOOKUP

    if reset_color_cycle:
        _reset_color_iterator()

    # get mesh positions and step sizes, cut in multi-materials
    pos = to_np(struct.positions)
    step = to_np(struct.step)

    if type(show) == str:
        if show.lower() == "auto" and pl is None:
            show = True
        else:
            False

    if pl is None:
        pl = pv.Plotter()

    if color == "auto":
        # colors = COLORS_DEFAULT
        diff_mat_names = [s.__name__ for s in struct.materials]
        mat_names = np.array(diff_mat_names)

        different_materials = np.unique(mat_names)

        mat_pos_subset_idx = []
        for pos_single_mat in different_materials:
            mat_pos_subset_idx.append(
                np.arange(len(mat_names))[mat_names == pos_single_mat]
            )
    else:
        different_materials = ["struct. id:{}".format(struct.id)]
        mat_pos_subset_idx = [np.arange(len(pos))]  # all pos
        # colors = [color]

    # the actual plot
    mesh_list = []
    for i_s, pos_idx in enumerate(mat_pos_subset_idx):
        pos_mat = pos[pos_idx]
        steplist_mat = step[pos_idx]

        pts = pv.PolyData(pos_mat)
        pts["steps"] = steplist_mat
        pts.set_active_scalars("steps")

        mesh_list.append(
            pts.glyph(geom=pv.Cube(), scale="steps", factor=scale, orient=False)
        )

    for i_s, mesh in enumerate(mesh_list):
        # chose color
        if color == "auto":
            # if material already used, re-use its color. otherwise use next color
            if different_materials[i_s] in LEGEND_ENTRIES_LOOKUP:
                _col = LEGEND_ENTRIES_LOOKUP[different_materials[i_s]]["color"]
            else:
                _col = next(COLOR_ITERATOR)
        elif type(color) in [list, tuple]:
            _col = color[i_s]
        else:
            _col = color

        # add legend entry:
        LEGEND_ENTRIES_LOOKUP[different_materials[i_s]] = dict(
            color=_col,
            fc=_col,
            ec=_col,
            marker="s",
            markersize=10,
        )

        label = different_materials[i_s]
        pl.add_mesh(
            mesh,
            color=_col,
            show_edges=show_grid,
            edge_color="black",
            line_width=0.5,
            opacity=alpha,
            edge_opacity=alpha,
            label=label,
        )

    if legend:
        _generate_legend(pl)

    if show:
        pl.show()

    return mesh_list


def _plot_structure_eff_3dpola(
    struct,
    scale=1.0,
    center_marker_scale=10,
    color="auto",
    sphere_style="wireframe",
    color_sphere="auto",
    theta_resolution=20,
    phi_resolution=20,
    alpha=0.1,
    show_grid=True,
    color_grid="auto",
    alpha_grid=0.05,
    show="auto",
    pl=None,
    legend=False,
    reset_color_cycle=True,
    **kwargs,
):
    from torchgdm import tools
    import numpy as np
    import pyvista as pv

    if "projection" in kwargs:
        kwargs.pop("projection")

    global COLOR_ITERATOR, LEGEND_ENTRIES_LOOKUP

    if reset_color_cycle:
        _reset_color_iterator()

    if type(show) == str:
        if show.lower() == "auto" and pl is None:
            show = True
        else:
            show = False

    if pl is None:
        pl = pv.Plotter()

    # legend label based on structure ID
    legend_label = "eff.pola. (id: {})".format(struct.id)

    # automatic next color
    if color == "auto":
        if legend_label in LEGEND_ENTRIES_LOOKUP:
            _col = LEGEND_ENTRIES_LOOKUP[legend_label]["color"]
        else:
            _col = next(COLOR_ITERATOR)
    else:
        _col = color

    if color_grid == "auto":
        color_grid = _col
    if color_sphere == "auto":
        color_sphere = _col

    # add / update legend entry
    LEGEND_ENTRIES_LOOKUP[legend_label] = dict(
        color=_col,
        fc=color_sphere,
        ec=_col,
        marker="o",
        markersize=10,
    )

    # geometry data to numpy
    pos_a = to_np(struct.positions)
    enclosing_radius = to_np(struct.step) / 2.0

    # create the actual plot: iterate over polarizabilities
    mesh_sphere_list = []
    mesh_center_list = []
    mesh_fullgeo_list = []
    for i, pos in enumerate(pos_a):
        r = enclosing_radius[i]
        _geo = struct.full_geometries[i]
        pos_mesh = to_np(_geo)
        step_mesh = to_np(tools.geometry.get_step_from_geometry(_geo))

        # plot enclosing sphere
        enclose_sphere = pv.Sphere(
            r + step_mesh / 2.0,
            pos,
            theta_resolution=theta_resolution,
            phi_resolution=phi_resolution,
        )
        mesh_sphere_list.append(enclose_sphere)

        # center pos. "marker" sphere
        mesh_center_list.append(pv.Sphere(center_marker_scale, pos))

        # full geometry mesh
        pts = pv.PolyData(pos_mesh)
        pts["steps"] = np.ones(len(pos_mesh)) * step_mesh
        pts.set_active_scalars("steps")

        mesh_fullgeo_list.append(
            pts.glyph(geom=pv.Cube(), scale="steps", factor=scale, orient=False)
        )

    # plot enclosing sphere wireframe
    for i_s, mesh in enumerate(mesh_sphere_list):
        pl.add_mesh(
            mesh,
            color=color_sphere,
            show_edges=False,
            line_width=0.5,
            edge_opacity=alpha,
            opacity=alpha,
            style=sphere_style,
        )

    # plot dipole position
    for i_s, mesh in enumerate(mesh_center_list):
        pl.add_mesh(mesh, color=_col)

    # optionally plot the replaced full geometry mesh
    if show_grid:
        for i_s, mesh in enumerate(mesh_fullgeo_list):
            pl.add_mesh(
                mesh,
                color=color_grid,
                show_edges=True,
                edge_color="black",
                line_width=0.5,
                opacity=alpha_grid,
                edge_opacity=alpha_grid * 0.1,
            )

    # optional legend additions
    if legend:
        _generate_legend(pl)

    if show:
        pl.show()

    return mesh_sphere_list, mesh_center_list, mesh_fullgeo_list


def structure(
    struct,
    color="auto",
    scale=1,
    legend=True,
    reset_color_cycle=True,
    pl=None,
    show="auto",
    **kwargs,
):
    """plot structure in 3D

    plot the structure `struct` in 3D using `pyvista`.
    Either from a structure instance, or using a simulation as input.

    kwargs are passed to individual structure plotting and / or to pyvista

    Parameters
    ----------
    struct : simulation or structure
          either a simulation or a structure instance

    projection : str, default: 'auto'
        which 2D projection to plot: "auto", "XY", "YZ", "XZ"

    color : str or matplotlib color, default: "auto"
            Color of scatterplot. Either "auto", or matplotlib-compatible color.
            "auto": automatic color selection (multi-color for multiple materials).

    legend : bool, default: True
        whether to add a legend if multi-material structure (requires auto-color enabled)


    Returns
    -------
        lists of shown meshes that are returned by the 3D plotting functions

    """
    import pyvista as pv
    from torchgdm.simulation import SimulationBase, Simulation
    from torchgdm.struct.volume.pola import StructDiscretized3D
    from torchgdm.struct.volume.pola import StructDiscretizedCubic3D
    from torchgdm.struct.volume.pola import StructDiscretizedHexagonal3D
    from torchgdm.struct.point.pola import StructEffPola3D

    if reset_color_cycle:
        _reset_color_iterator()

    # got a structure instance:
    if (
        issubclass(type(struct), StructDiscretized3D)
        or issubclass(type(struct), StructDiscretizedCubic3D)
        or issubclass(type(struct), StructDiscretizedHexagonal3D)
    ):
        pl_res = _plot_structure_discretized(
            struct,
            color=color,
            scale=scale,
            show=show,
            pl=pl,
            **kwargs,
        )
    elif issubclass(type(struct), StructEffPola3D):
        pl_res = _plot_structure_eff_3dpola(
            struct,
            color=color,
            scale=scale,
            show=show,
            pl=pl,
            **kwargs,
        )
    elif issubclass(type(struct), Simulation) or issubclass(
        type(struct), SimulationBase
    ):
        # -- prep
        sim = struct

        if type(show) == str:
            if show.lower() == "auto" and pl is None:
                show = True
            else:
                show = False

        if pl is None:
            pl = pv.Plotter()

        # -- call all structure's plot functions
        pl_res = []  # collect results
        for i_s, _st in enumerate(sim.structures):
            pl_res.append(
                _st.plot3d(
                    color=color,
                    scale=scale,
                    pl=pl,
                    legend=False,
                    reset_color_cycle=False,
                    show=False,
                    **kwargs,
                )
            )

        # -- finalize: config global plot
        if legend:
            _generate_legend(pl)
        if show:
            pl.show()
    else:
        raise ValueError("Unknown structure input")

    if reset_color_cycle:
        _reset_color_iterator()

    return pl_res



# -*- coding: utf-8 -*-
"""package for materials optical properties
"""
from .hardcoded import MatConstant
from .hardcoded import MatTiO2

from .tabulated import list_available_materials
from .tabulated import MatDatabase



# -*- coding: utf-8 -*-
"""2D visualization tools based on matplotlib

requirements:
 - numpy
 - matplotlib
 - scipy (for scalar fields)
 
"""
from . import geo2d, vec2d, scalar2d

from .geo2d import _reset_color_iterator

from .geo2d import structure, contour
from .vec2d import vectorfield, vectorfield_inside, streamlines_energy_flux
from .scalar2d import field_intensity, field_amplitude
from .scalar2d import _scalarfield as scalarfield
